# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module containing variant scorer dataclasses for variant scoring."""


from collections.abc import Sequence
import dataclasses
import enum
import itertools
from typing import TypeVar

from alphagenome.models import dna_output
from alphagenome.protos import dna_model_pb2
import anndata
import immutabledict
import pandas as pd


class AggregationType(enum.Enum):
  """Enum indicating the type of variant scorer aggregation.

  Attributes:
    DIFF_MEAN: Difference of means, i.e., mean(`ALT`) - mean(`REF`).
    DIFF_SUM: Difference of sums, i.e., sum(`ALT`) - sum(`REF`).
    DIFF_SUM_LOG2: Log scales predictions, then takes the sum and then the
      difference, i.e., sum(log2(`ALT`)) - sum(log2(`REF`)).
    DIFF_LOG2_SUM: Takes the sum of predictions, applies a log transform, then
      takes the difference between predictions, i.e., log2(sum(`ALT`)) -
      log2(sum(`REF`)).
    L2_DIFF: Takes the difference of `ALT` and `REF` predictions, then computes
      the L2 norm, i.e., l2_norm(`ALT` - `REF`).
    L2_DIFF_LOG1P: Log scales the predictions + 1, takes the difference of `ALT`
      and `REF` predictions, then computes the L2 norm, i.e.,
      l2_norm(log1p(`ALT`) - log1p(`REF`)).
    ACTIVE_MEAN: Maximum of means, i.e., max(mean(`ALT`), mean(`REF`)).
    ACTIVE_SUM: Maximum of sums, i.e., max(sum(`ALT), sum(`REF`)).

  Methods:
    to_proto: Converts the aggregation type to its corresponding proto enum.
  """

  DIFF_MEAN = dna_model_pb2.AggregationType.AGGREGATION_TYPE_DIFF_MEAN
  DIFF_SUM = dna_model_pb2.AggregationType.AGGREGATION_TYPE_DIFF_SUM
  DIFF_SUM_LOG2 = dna_model_pb2.AggregationType.AGGREGATION_TYPE_DIFF_SUM_LOG2
  DIFF_LOG2_SUM = dna_model_pb2.AggregationType.AGGREGATION_TYPE_DIFF_LOG2_SUM
  L2_DIFF = dna_model_pb2.AggregationType.AGGREGATION_TYPE_L2_DIFF
  L2_DIFF_LOG1P = dna_model_pb2.AggregationType.AGGREGATION_TYPE_L2_DIFF_LOG1P
  ACTIVE_MEAN = dna_model_pb2.AggregationType.AGGREGATION_TYPE_ACTIVE_MEAN
  ACTIVE_SUM = dna_model_pb2.AggregationType.AGGREGATION_TYPE_ACTIVE_SUM

  def to_proto(self) -> dna_model_pb2.AggregationType:
    return self.value

  def __repr__(self) -> str:
    return self.name


class BaseVariantScorer(enum.Enum):
  """Enum indicating the type of variant scoring.

  Attributes:
    CENTER_MASK: Center mask scorer.
    CONTACT_MAP: Contact map scorer, a center mask scorer that handles 2D
      tensors.
    GENE_MASK_LFC: Gene-mask scoring based on log fold change.
    GENE_MASK_ACTIVE: Gene-mask scoring based on active allele.
    GENE_MASK_SPLICING: Gene-mask scoring for splicing.
    PA_QTL: Polyadenylation QTL scoring.
    SPLICE_JUNCTION: Splice junction scoring.
  """

  CENTER_MASK = enum.auto()
  CONTACT_MAP = enum.auto()
  GENE_MASK_LFC = enum.auto()
  GENE_MASK_ACTIVE = enum.auto()
  GENE_MASK_SPLICING = enum.auto()
  PA_QTL = enum.auto()
  SPLICE_JUNCTION = enum.auto()


SUPPORTED_ORGANISMS = immutabledict.immutabledict({
    BaseVariantScorer.CENTER_MASK: dna_model_pb2.Organism.values(),
    BaseVariantScorer.CONTACT_MAP: dna_model_pb2.Organism.values(),
    BaseVariantScorer.GENE_MASK_LFC: dna_model_pb2.Organism.values(),
    BaseVariantScorer.GENE_MASK_ACTIVE: dna_model_pb2.Organism.values(),
    BaseVariantScorer.GENE_MASK_SPLICING: dna_model_pb2.Organism.values(),
    BaseVariantScorer.PA_QTL: [dna_model_pb2.Organism.ORGANISM_HOMO_SAPIENS],
    BaseVariantScorer.SPLICE_JUNCTION: dna_model_pb2.Organism.values(),
})

SUPPORTED_OUTPUT_TYPES = immutabledict.immutabledict({
    BaseVariantScorer.CENTER_MASK: [
        dna_output.OutputType.ATAC,
        dna_output.OutputType.CAGE,
        dna_output.OutputType.DNASE,
        dna_output.OutputType.PROCAP,
        dna_output.OutputType.RNA_SEQ,
        dna_output.OutputType.CHIP_HISTONE,
        dna_output.OutputType.CHIP_TF,
        dna_output.OutputType.SPLICE_SITES,
        dna_output.OutputType.SPLICE_SITE_USAGE,
    ],
    BaseVariantScorer.CONTACT_MAP: [dna_output.OutputType.CONTACT_MAPS],
    BaseVariantScorer.GENE_MASK_LFC: [
        dna_output.OutputType.ATAC,
        dna_output.OutputType.CAGE,
        dna_output.OutputType.DNASE,
        dna_output.OutputType.PROCAP,
        dna_output.OutputType.RNA_SEQ,
        dna_output.OutputType.SPLICE_SITES,
        dna_output.OutputType.SPLICE_SITE_USAGE,
    ],
    BaseVariantScorer.GENE_MASK_ACTIVE: [
        dna_output.OutputType.ATAC,
        dna_output.OutputType.CAGE,
        dna_output.OutputType.DNASE,
        dna_output.OutputType.PROCAP,
        dna_output.OutputType.RNA_SEQ,
        dna_output.OutputType.SPLICE_SITES,
        dna_output.OutputType.SPLICE_SITE_USAGE,
    ],
    BaseVariantScorer.GENE_MASK_SPLICING: [
        dna_output.OutputType.SPLICE_SITES,
        dna_output.OutputType.SPLICE_SITE_USAGE,
    ],
})

SUPPORTED_WIDTHS = immutabledict.immutabledict({
    BaseVariantScorer.CENTER_MASK: [None, 501, 2001, 10_001, 100_001, 200_001],
    BaseVariantScorer.GENE_MASK_SPLICING: [None, 101, 1_001, 10_001],
})

SUPPORTED_AGGREGATIONS = immutabledict.immutabledict({
    BaseVariantScorer.CENTER_MASK: [
        AggregationType.DIFF_MEAN,
        AggregationType.DIFF_SUM,
        AggregationType.DIFF_SUM_LOG2,
        AggregationType.DIFF_LOG2_SUM,
        AggregationType.L2_DIFF,
        AggregationType.L2_DIFF_LOG1P,
        AggregationType.ACTIVE_MEAN,
        AggregationType.ACTIVE_SUM,
    ],
})


@dataclasses.dataclass(frozen=True)
class CenterMaskScorer:
  """Variant scorer for center mask scorers.

  Variant scorer that aggregates ALT and REF values using a spatial mask
  centered around the variant before computing the difference. This scorer
  aggregates over the spatial axis and returns one score per output track.

  Attributes:
    requested_output: The requested output type (e.g., ATAC, DNASE, etc.)
    width: The width of the mask around the variant. If None, the score is
      computed over the entire sequence.
    aggregation_type: The aggregation type.
    base_variant_scorer: The base variant scorer.
    name: The name of the scorer (a composite of the above attributes that
      uniquely identifies a scorer combination).
    is_signed: Whether this variant scorer is directional (such that scores may
      be negative or positive) or non-directional (scores are always positive).

  Methods:
    to_proto: Converts the scorer to its corresponding proto message.

  Raises:
    ValueError: If the requested output, width, or aggregation type is not
      supported.
  """

  requested_output: dna_output.OutputType
  width: int | None
  aggregation_type: AggregationType

  @property
  def base_variant_scorer(self) -> BaseVariantScorer:
    return BaseVariantScorer.CENTER_MASK

  @property
  def name(self) -> str:
    return str(self)

  @property
  def is_signed(self) -> bool:
    # The 'active' scorers track the maximum of the ALT and REF, and are
    # non-directional. All other scorers track ALT - REF, with L2_DIFF
    # nondirectional and the remaining directional.
    return self.aggregation_type in [
        AggregationType.DIFF_MEAN,
        AggregationType.DIFF_SUM,
        AggregationType.DIFF_SUM_LOG2,
        AggregationType.DIFF_LOG2_SUM,
    ]

  def __post_init__(self):
    if (
        self.requested_output
        not in SUPPORTED_OUTPUT_TYPES[self.base_variant_scorer]
    ):
      raise ValueError(
          f'Unsupported requested output: {self.requested_output}. Supported'
          f' output types: {SUPPORTED_OUTPUT_TYPES[self.base_variant_scorer]}'
      )
    if self.width not in SUPPORTED_WIDTHS[self.base_variant_scorer]:
      raise ValueError(
          f'Unsupported width: {self.width}. Supported widths:'
          f' {SUPPORTED_WIDTHS[self.base_variant_scorer]}'
      )
    if (
        self.aggregation_type
        not in SUPPORTED_AGGREGATIONS[self.base_variant_scorer]
    ):
      raise ValueError(
          f'Unsupported aggregation type: {self.aggregation_type}. Supported'
          ' aggregation types:'
          f' {SUPPORTED_AGGREGATIONS[self.base_variant_scorer]}'
      )

  def to_proto(self) -> dna_model_pb2.VariantScorer:
    return dna_model_pb2.VariantScorer(
        center_mask=dna_model_pb2.CenterMaskScorer(
            requested_output=self.requested_output.to_proto(),
            width=self.width,
            aggregation_type=self.aggregation_type.to_proto(),
        )
    )


@dataclasses.dataclass(frozen=True)
class ContactMapScorer:
  """Variant scorer for contact map scorers.

  Variant scorer that quantifies local contact disruption between ALT and REF
  alleles. Uses a 1MB window centered around the variant to calculate the
  mean absolute difference of contact frequencies, for all interactions
  involving the variant-containing bin.

  Attributes:
    requested_output: The requested output type (e.g., CONTACT_MAPS).
    base_variant_scorer: The base variant scorer.
    name: The name of the scorer (a composite of the above attributes that
      uniquely identifies a scorer combination).
    is_signed: Whether this variant scorer is directional (such that scores may
      be negative or positive) or non-directional (scores are always positive).

  Methods:
    to_proto: Converts the scorer to its corresponding proto message.
  """

  @property
  def base_variant_scorer(self) -> BaseVariantScorer:
    return BaseVariantScorer.CONTACT_MAP

  @property
  def name(self) -> str:
    return str(self)

  @property
  def is_signed(self) -> bool:
    return False

  def to_proto(self) -> dna_model_pb2.VariantScorer:
    return dna_model_pb2.VariantScorer(
        contact_map=dna_model_pb2.ContactMapScorer()
    )

  @property
  def requested_output(self) -> dna_output.OutputType:
    return dna_output.OutputType.CONTACT_MAPS


@dataclasses.dataclass(frozen=True)
class GeneMaskLFCScorer:
  """Variant scorer for gene-mask log fold change scoring.

  Variant scorer that quantifies impact on overall gene transcript abundance.
  Calculates the log fold change of gene expression level between ALT and REF
  alleles using a gene exon mask.

  Attributes:
    requested_output: The requested output type (e.g. ATAC, DNASE, etc.)
    base_variant_scorer: The base variant scorer.
    name: The name of the scorer (a composite of the above attributes that
      uniquely identifies a scorer combination).
    is_signed: Whether this variant scorer is directional (such that scores may
      be negative or positive) or non-directional (scores are always positive).

  Methods:
    to_proto: Converts the scorer to its corresponding proto message.

  Raises:
    ValueError: If the requested output is not supported.
  """

  requested_output: dna_output.OutputType

  @property
  def base_variant_scorer(self) -> BaseVariantScorer:
    return BaseVariantScorer.GENE_MASK_LFC

  @property
  def name(self) -> str:
    return str(self)

  @property
  def is_signed(self) -> bool:
    return True

  def __post_init__(self):
    if (
        self.requested_output
        not in SUPPORTED_OUTPUT_TYPES[self.base_variant_scorer]
    ):
      raise ValueError(
          f'Unsupported requested output: {self.requested_output}. Supported'
          f' output types: {SUPPORTED_OUTPUT_TYPES[self.base_variant_scorer]}'
      )

  def to_proto(self) -> dna_model_pb2.VariantScorer:
    return dna_model_pb2.VariantScorer(
        gene_mask=dna_model_pb2.GeneMaskLFCScorer(
            requested_output=self.requested_output.to_proto(),
        )
    )


@dataclasses.dataclass(frozen=True)
class GeneMaskActiveScorer:
  """Variant scorer for gene-mask active allele scoring.

  Variant scorer that captures absolute activity level associated with one of
  the alleles, rather than quantifying the difference between the ALT and REF.
  Active allele gene scores are calculated by taking the maximum of the
  aggregated ALT and REF signals across exons for a gene of interest.

  Attributes:
    requested_output: The requested output type (e.g. ATAC, DNASE, etc.)
    base_variant_scorer: The base variant scorer.
    name: The name of the scorer (a composite of the above attributes that
      uniquely identifies a scorer combination).
    is_signed: Whether this variant scorer is directional (such that scores may
      be negative or positive) or non-directional (scores are always positive).

  Methods:
    to_proto: Converts the scorer to its corresponding proto message.

  Raises:
    ValueError: If the requested output is not supported.
  """

  requested_output: dna_output.OutputType

  @property
  def base_variant_scorer(self) -> BaseVariantScorer:
    return BaseVariantScorer.GENE_MASK_ACTIVE

  @property
  def name(self) -> str:
    return str(self)

  @property
  def is_signed(self) -> bool:
    return False

  def __post_init__(self):
    if (
        self.requested_output
        not in SUPPORTED_OUTPUT_TYPES[self.base_variant_scorer]
    ):
      raise ValueError(
          f'Unsupported requested output: {self.requested_output}. Supported'
          f' output types: {SUPPORTED_OUTPUT_TYPES[self.base_variant_scorer]}'
      )

  def to_proto(self) -> dna_model_pb2.VariantScorer:
    return dna_model_pb2.VariantScorer(
        gene_mask_active=dna_model_pb2.GeneMaskActiveScorer(
            requested_output=self.requested_output.to_proto(),
        )
    )


@dataclasses.dataclass(frozen=True)
class GeneMaskSplicingScorer:
  """Variant scorer for gene-mask scoring for splicing.

  Variant scorer that quantifies changes in class assignment probabilties
  (dna_output.OutputType.SPLICE_SITES) or changes in the usage of splice sites
  (dna_output.OutputType.SPLICE_SITE_USAGE) between ALT and REF alleles using
  a gene exon mask.

  Attributes:
    requested_output: The requested output type (e.g. ATAC, DNASE, etc.)
    width: The width of the mask around the variant.
    base_variant_scorer: The base variant scorer.
    name: The name of the scorer (a composite of the above attributes that
      uniquely identifies a scorer combination).
    is_signed: Whether this variant scorer is directional (such that scores may
      be negative or positive) or non-directional (scores are always positive).

  Methods:
    to_proto: Converts the scorer to its corresponding proto message.

  Raises:
    ValueError: If the requested output or width is not supported.
  """

  requested_output: dna_output.OutputType
  width: int | None

  @property
  def base_variant_scorer(self) -> BaseVariantScorer:
    return BaseVariantScorer.GENE_MASK_SPLICING

  @property
  def name(self) -> str:
    return str(self)

  @property
  def is_signed(self) -> bool:
    return False

  def __post_init__(self):
    if (
        self.requested_output
        not in SUPPORTED_OUTPUT_TYPES[self.base_variant_scorer]
    ):
      raise ValueError(
          f'Unsupported requested output: {self.requested_output}. Supported'
          f' output types: {SUPPORTED_OUTPUT_TYPES[self.base_variant_scorer]}'
      )
    if self.width not in SUPPORTED_WIDTHS[self.base_variant_scorer]:
      raise ValueError(
          f'Unsupported width: {self.width}. Supported widths:'
          f' {SUPPORTED_WIDTHS[self.base_variant_scorer]}'
      )

  def to_proto(self) -> dna_model_pb2.VariantScorer:
    return dna_model_pb2.VariantScorer(
        gene_mask_splicing=dna_model_pb2.GeneMaskSplicingScorer(
            requested_output=self.requested_output.to_proto(),
            width=self.width,
        )
    )


@dataclasses.dataclass(frozen=True)
class PolyadenylationScorer:
  """Variant scorer for polyadenylation QTLs (paQTLs).

  Variant scorer for polyadenylation quantitative trait loci (paQTLs) to capture
  a variant's impact on RNA isoform production.

  The scoring approach quantifies the maximum log fold change in expression
  between the set of proximal polyadenylation sites (PAS)s vs. the set of
  distal PASs, where the sets of proximal and distal PASs are the PASs upstream
  or downstream of any given 3' cleavage site respectively.

  Attributes:
    base_variant_scorer: The base variant scorer.
    name: The name of the scorer (a composite of the above attributes that
      uniquely identifies a scorer combination).
    is_signed: Whether this variant scorer is directional (such that scores may
      be negative or positive) or non-directional (scores are always positive).

  Methods:
    to_proto: Converts the scorer to its corresponding proto message.
  """

  @property
  def requested_output(self) -> dna_output.OutputType:
    """Returns the scorer's underlying output type."""
    return dna_output.OutputType.RNA_SEQ

  @property
  def base_variant_scorer(self) -> BaseVariantScorer:
    return BaseVariantScorer.PA_QTL

  @property
  def name(self) -> str:
    return str(self)

  @property
  def is_signed(self) -> bool:
    return False

  def to_proto(self) -> dna_model_pb2.VariantScorer:
    return dna_model_pb2.VariantScorer(
        pa_qtl=dna_model_pb2.PolyadenylationScorer()
    )


@dataclasses.dataclass(frozen=True)
class SpliceJunctionScorer:
  """Variant scorer for splice junction scoring.

  Attributes:
    base_variant_scorer: The base variant scorer.
    name: The name of the scorer (a composite of the above attributes that
      uniquely identifies a scorer combination).
    is_signed: Whether this variant scorer is directional (such that scores may
      be negative or positive) or non-directional (scores are always positive).

  Methods:
    to_proto: Converts the scorer to its corresponding proto message.
  """

  @property
  def requested_output(self) -> dna_output.OutputType:
    """Returns the scorer's underlying output type."""
    return dna_output.OutputType.SPLICE_JUNCTIONS

  @property
  def base_variant_scorer(self) -> BaseVariantScorer:
    return BaseVariantScorer.SPLICE_JUNCTION

  @property
  def name(self) -> str:
    return str(self)

  @property
  def is_signed(self) -> bool:
    return False

  def to_proto(self) -> dna_model_pb2.VariantScorer:
    return dna_model_pb2.VariantScorer(
        splice_junction=dna_model_pb2.SpliceJunctionScorer()
    )


# TypeVar for all variant scorer types.
VariantScorerTypes = TypeVar(
    'VariantScorerTypes',
    CenterMaskScorer,
    ContactMapScorer,
    GeneMaskLFCScorer,
    GeneMaskActiveScorer,
    GeneMaskSplicingScorer,
    PolyadenylationScorer,
    SpliceJunctionScorer,
)

# A dict of variant scorers, with our recommended settings for a wide range
# of different use cases and settings.
RECOMMENDED_VARIANT_SCORERS = immutabledict.immutabledict({
    'ATAC': CenterMaskScorer(
        requested_output=dna_output.OutputType.ATAC,
        width=501,
        aggregation_type=AggregationType.DIFF_LOG2_SUM,
    ),
    'CONTACT_MAPS': ContactMapScorer(),
    'DNASE': CenterMaskScorer(
        requested_output=dna_output.OutputType.DNASE,
        width=501,
        aggregation_type=AggregationType.DIFF_LOG2_SUM,
    ),
    'CHIP_TF': CenterMaskScorer(
        requested_output=dna_output.OutputType.CHIP_TF,
        width=501,
        aggregation_type=AggregationType.DIFF_LOG2_SUM,
    ),
    'CHIP_HISTONE': CenterMaskScorer(
        requested_output=dna_output.OutputType.CHIP_HISTONE,
        width=2001,
        aggregation_type=AggregationType.DIFF_LOG2_SUM,
    ),
    'CAGE': CenterMaskScorer(
        requested_output=dna_output.OutputType.CAGE,
        width=501,
        aggregation_type=AggregationType.DIFF_LOG2_SUM,
    ),
    'PROCAP': CenterMaskScorer(
        requested_output=dna_output.OutputType.PROCAP,
        width=501,
        aggregation_type=AggregationType.DIFF_LOG2_SUM,
    ),
    # Expression log fold change.
    'RNA_SEQ': GeneMaskLFCScorer(
        requested_output=dna_output.OutputType.RNA_SEQ
    ),
    'RNA_SEQ_ACTIVE': GeneMaskActiveScorer(
        requested_output=dna_output.OutputType.RNA_SEQ
    ),
    'SPLICE_SITES': GeneMaskSplicingScorer(
        requested_output=dna_output.OutputType.SPLICE_SITES,
        width=None,
    ),
    'SPLICE_SITE_USAGE': GeneMaskSplicingScorer(
        requested_output=dna_output.OutputType.SPLICE_SITE_USAGE,
        width=None,
    ),
    'SPLICE_JUNCTIONS': SpliceJunctionScorer(),
    'POLYADENYLATION': PolyadenylationScorer(),
    'ATAC_ACTIVE': CenterMaskScorer(
        requested_output=dna_output.OutputType.ATAC,
        width=501,
        aggregation_type=AggregationType.ACTIVE_SUM,
    ),
    'DNASE_ACTIVE': CenterMaskScorer(
        requested_output=dna_output.OutputType.DNASE,
        width=501,
        aggregation_type=AggregationType.ACTIVE_SUM,
    ),
    'CHIP_TF_ACTIVE': CenterMaskScorer(
        requested_output=dna_output.OutputType.CHIP_TF,
        width=501,
        aggregation_type=AggregationType.ACTIVE_SUM,
    ),
    'CHIP_HISTONE_ACTIVE': CenterMaskScorer(
        requested_output=dna_output.OutputType.CHIP_HISTONE,
        width=2001,
        aggregation_type=AggregationType.ACTIVE_SUM,
    ),
    'CAGE_ACTIVE': CenterMaskScorer(
        requested_output=dna_output.OutputType.CAGE,
        width=501,
        aggregation_type=AggregationType.ACTIVE_SUM,
    ),
    'PROCAP_ACTIVE': CenterMaskScorer(
        requested_output=dna_output.OutputType.PROCAP,
        width=501,
        aggregation_type=AggregationType.ACTIVE_SUM,
    ),
})


def get_recommended_scorers(
    organism: dna_model_pb2.Organism,
) -> list[VariantScorerTypes]:
  """Returns the recommended variant scorers for a given organism."""
  return [
      scorer
      for scorer in RECOMMENDED_VARIANT_SCORERS.values()
      if organism in SUPPORTED_ORGANISMS[scorer.base_variant_scorer]
  ]


def tidy_anndata(
    adata: anndata.AnnData,
    match_gene_strand: bool = True,
    include_extended_metadata: bool = True,
) -> pd.DataFrame:
  """Formats an :class:`~anndata.AnnData` score as a tidy DataFrame.

  This function converts the score output from an AnnData object into a
  long-format pandas DataFrame, where each row represents:

    - For non-gene-centric variant scorers: Score for a variant-track pair.
    - For non-gene-centric interval scorers: Score for an interval-track pair.
    - For gene-centric variant/interval scoring: Score for a
      variant/interval-gene-track combination.

  Args:
    adata: An AnnData object containing scores.
    match_gene_strand: If True (and using gene-centric scoring), rows with
      mismatched gene and track strands are removed.
    include_extended_metadata: If True, includes additional columns derived from
      metadata specific to the output type, such as biosample name and type,
      gtex tissue, transcription factor, and histone mark, if available. If
      False, only includes minimal metadata columns required to unique identify
      a track withing a given output type: track_name and track_strand.

  Returns:
    A pandas DataFrame with one score per row. The DataFrame includes
    columns for variant ID (if applicable), scored interval, gene information
    (if applicable), output type, variant/interval scorer, track name,
    ontology term, assay type, track strand, and raw score. Additional metadata
    such as biosample name and type, gtex tissue are also returned (where
    available). See :func:`full_path_to.tidy_scores` for more details on the
    returned columns.

  Raises:
    ValueError: If the input is not an AnnData object.
  """
  if not isinstance(adata, anndata.AnnData):
    raise ValueError('Invalid input type. Must be an AnnData object.')

  # Columns to include from the gene metadata. If the column did not
  # exist in the original metadata (or if the scores do not have a concept
  # of a gene), a column of all Nones is added.
  gene_columns = [
      'gene_id',
      'gene_name',
      'gene_type',
      'gene_strand',
      'junction_Start',
      'junction_End',
  ]
  if 'gene_id' in adata.obs and 'strand' in adata.obs:
    # Scores are for a gene-based scorer.
    obs = adata.obs.rename({'strand': 'gene_strand'}, axis=1)
    obs['gene_id'] = obs['gene_id'].str.split('.', expand=True)[0]  # Depatch.
    for col in gene_columns:
      if col not in obs:
        obs[col] = None
  elif adata.X.shape[0] == 0:
    # Scores are empty, so we return an empty dataframe.
    return pd.DataFrame()
  elif adata.X.shape[0] == 1 and adata.obs.empty:
    # Scores are for a non-gene-based scorer.
    obs = pd.DataFrame([[None] * len(gene_columns)], columns=gene_columns)
  else:
    raise ValueError('Scores contain gene metadata but no gene_id/strand.')

  # Columns to include from the track metadata. Only added if the column
  # appears in the original metadata.
  var = adata.var.rename(
      {'strand': 'track_strand', 'name': 'track_name'}, axis=1
  )
  track_columns = ['track_name', 'track_strand']
  if include_extended_metadata:
    extended_metadata_columns = [
        'Assay title',
        'ontology_curie',
        'biosample_name',
        'biosample_type',
        'gtex_tissue',
        'transcription_factor',
        'histone_mark',
    ]
    track_columns.extend([c for c in extended_metadata_columns if c in var])

  # Construct score dataframe.
  df = pd.merge(var, obs, how='cross')
  df['scored_interval'] = adata.uns['interval']

  # Add id and scorer information depending on the type of score.
  if 'variant_scorer' in adata.uns:
    df['variant_id'] = adata.uns['variant']
    df['output_type'] = adata.uns['variant_scorer'].requested_output.name
    df['variant_scorer'] = str(adata.uns['variant_scorer'])
    id_columns = ['variant_id', 'scored_interval']
    scorer_columns = ['output_type', 'variant_scorer']
  elif 'interval_scorer' in adata.uns:
    df['output_type'] = adata.uns['interval_scorer'].requested_output.name
    df['interval_scorer'] = str(adata.uns['interval_scorer'])
    id_columns = ['scored_interval']
    scorer_columns = ['output_type', 'interval_scorer']
  else:
    raise ValueError(
        'Invalid scores, expected either variant_scorer or interval_scorer.'
    )

  # Formatting final touches.
  df = df.loc[:, id_columns + gene_columns + scorer_columns + track_columns]
  df['raw_score'] = adata.X.T.reshape((-1,))
  if 'quantiles' in adata.layers:
    df['quantile_score'] = adata.layers['quantiles'].T.reshape((-1,))

  # Remove entries where DNA strands are mismatched. Note that we still retain
  # all tracks that have strand '.' which indicates the data is unstranded.
  if match_gene_strand:
    mismatched_strands = (df['gene_strand'] == '+') & (
        df['track_strand'] == '-'
    ) | (df['gene_strand'] == '-') & (df['track_strand'] == '+')
    df = df[~mismatched_strands].reset_index(drop=True)
  return df.reset_index(drop=True)


def tidy_scores(
    scores: Sequence[anndata.AnnData] | Sequence[Sequence[anndata.AnnData]],
    match_gene_strand: bool = True,
    include_extended_metadata: bool = True,
) -> pd.DataFrame | None:
  """Formats scores into a tidy (long) pandas DataFrame.

  This function reformats variant scores into a more readable DataFrame with one
  score per row. This function supports both scores generated from variant
  scorers (e.g., the output of `score_variant` or `score_variants`) or interval
  scorers (e.g., the output of `score_interval` or `score_intervals`).

  It handles both variant/interval-centric scoring (producing one score row per
  variant/interval-track pair) and gene-centric scoring (one score row per
  variant/interval-gene-track combination).

  The function accepts these score input types:

  - A sequence of AnnData objects (e.g., output of `score_variant` with one or
    more scorers).
  - A nested sequence of AnnData objects (e.g., output of `score_variants`
    with multiple variants).

  Scores from multiple scorers or multiple variants/intervals are concatenated
  together into a single pandas DataFrame, containing the union of all
  applicable columns across scorers.

  Args:
    scores: Scoring output as either a sequence of AnnData objects, or a nested
      sequence of AnnData objects (for example, the outputs of `score_variant`
      and `score_variants`, respectively).
    match_gene_strand: If True (and using gene-centric scoring), rows with
      mismatched gene and track strands are removed.
    include_extended_metadata: Argument passed to `tidy_anndata` to include
      additional metadata columns where available.

  Returns:
    pd.DataFrame with columns:

      - variant_id (when applicable): Variant of interest (e.g.
        chr22:36201698:A>C).
      - scored_interval: Genomic interval scored (e.g. chr22:36100000-36300000).
      - gene_id: ENSEMBL gene identifier without version number.
        (e.g. ENSG00000100342), or None if not applicable.
      - gene_name: HGNC gene symbol (e.g. APOL1), or None if not applicable.
      - gene_type: Gene biotype (e.g. protein_coding, lncRNA), or None if not
        applicable.
      - gene_strand: Strand of the gene ('+', '-', '.'), or None if not
        applicable.
      - output_type: Type of the output from the model (e.g. RNA_SEQ, DNASE).
      - interval_scorer (when applicable): Name of the interval scorer used.
      - variant_scorer (when applicable): Name of the variant scorer used.
      - track_name: Name of the output track (e.g. UBERON:0036149 total
        RNA-seq).
      - track_strand: Strand of the track ('+', '-', or '.').
      - ontology_curie: Ontology term for the cell type or tissue of the track
        (e.g. UBERON:0036149), or NaN if not applicable.
      - gtex_tissue: Name of the gtex tissue (e.g. Liver), or NaN if not
        applicable.
      - Assay title: Subtype of the assay (e.g., total RNA-seq), or NaN if not
        applicable.
      - biosample_name: Name of the biosample (e.g. liver), or NaN if not
        applicable.
      - biosample_type: Type of biosample (e.g. 'tissue' or 'primary cell'), or
        NaN if not applicable.
      - transcription_factor: Name of the transcription factor (e.g. 'CTCF'), or
        NaN if not applicable.
      - histone_mark: Name of the histological mark (e.g. 'H3K4ME3'), or NaN if
        not applicable.
      - raw_score: Raw variant score.
      - quantile_score (when applicable): Quantile score.

  Raises:
    ValueError: If the input is not a valid type (sequence of AnnData or nested
      sequence of AnnDatas).
  """
  if isinstance(scores, Sequence):
    tidied_anndata = [
        tidy_anndata(adata, match_gene_strand, include_extended_metadata)
        for adata in itertools.chain.from_iterable(scores)
    ]
    if not tidied_anndata:
      return None

    concatenated_df = pd.concat(tidied_anndata, axis=0, ignore_index=True)
    # Put the raw_score (and quantile_score when applicable) columns last.
    last_columns = ['raw_score']
    if 'quantile_score' in concatenated_df.columns:
      last_columns.append('quantile_score')
    return concatenated_df[
        [col for col in concatenated_df.columns if col not in last_columns]
        + last_columns
    ]
  else:
    raise ValueError(
        'Invalid input type. Must be sequence of AnnDatas or nested sequence of'
        ' AnnDatas.'
    )



# Copyright 2025 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions for Google Colab."""

import os


def get_api_key(secret: str = 'ALPHA_GENOME_API_KEY'):
  """Returns API key from environment variable or Colab secrets.

  Tries to retrieve the API key from the environment first. If not found,
  attempts to retrieve it from Colab secrets (if running in Colab).

  Args:
    secret: The name of the environment variable or Colab secret key to
      retrieve.

  Raises:
    ValueError: If the API key cannot be found in the environment or Colab
      secrets.
  """

  if api_key := os.environ.get(secret):
    return api_key

  try:
    # pylint: disable=g-import-not-at-top, import-outside-toplevel
    from google.colab import userdata  # pytype: disable=import-error
    # pylint: enable=g-import-not-at-top, import-outside-toplevel

    try:
      api_key = userdata.get(secret)
      return api_key
    except (
        userdata.NotebookAccessError,
        userdata.SecretNotFoundError,
        userdata.TimeoutException,
    ) as e:
      raise ValueError(
          f'Cannot find or access API key in Colab secrets with {secret=}. Make'
          ' sure you have added the API key to Colab secrets and enabled'
          ' access. See'
          ' https://www.alphagenomedocs.com/installation.html#add-api-key-to-secrets'
          ' for more details.'
      ) from e
  except ImportError:
    # Not running in Colab.
    pass

  raise ValueError(f'Cannot find API key with {secret=}.')



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest
from absl.testing import parameterized
from alphagenome.models import dna_output
from alphagenome.models import interval_scorers
from alphagenome.models import variant_scorers
from alphagenome.protos import dna_model_pb2
import anndata
import numpy as np
import pandas as pd


class VariantScorersTest(parameterized.TestCase):

  def test_aggregation_type_to_proto(self):
    expected = [
        dna_model_pb2.AggregationType.AGGREGATION_TYPE_DIFF_MEAN,
        dna_model_pb2.AggregationType.AGGREGATION_TYPE_DIFF_SUM,
        dna_model_pb2.AggregationType.AGGREGATION_TYPE_DIFF_SUM_LOG2,
        dna_model_pb2.AggregationType.AGGREGATION_TYPE_DIFF_LOG2_SUM,
        dna_model_pb2.AggregationType.AGGREGATION_TYPE_L2_DIFF,
        dna_model_pb2.AggregationType.AGGREGATION_TYPE_L2_DIFF_LOG1P,
        dna_model_pb2.AggregationType.AGGREGATION_TYPE_ACTIVE_MEAN,
        dna_model_pb2.AggregationType.AGGREGATION_TYPE_ACTIVE_SUM,
    ]

    for variant_scorer_aggregation_type, expected_proto in zip(
        variant_scorers.AggregationType, expected, strict=True
    ):
      self.assertEqual(
          variant_scorer_aggregation_type.to_proto(), expected_proto
      )

  def test_variant_scorers_to_proto(self):
    expected = [
        dna_model_pb2.VariantScorer(
            center_mask=dna_model_pb2.CenterMaskScorer(
                width=10_001,
                aggregation_type=(
                    dna_model_pb2.AggregationType.AGGREGATION_TYPE_DIFF_MEAN
                ),
                requested_output=dna_model_pb2.OUTPUT_TYPE_RNA_SEQ,
            )
        ),
        dna_model_pb2.VariantScorer(
            center_mask=dna_model_pb2.CenterMaskScorer(
                width=10_001,
                aggregation_type=(
                    dna_model_pb2.AggregationType.AGGREGATION_TYPE_ACTIVE_MEAN
                ),
                requested_output=dna_model_pb2.OUTPUT_TYPE_PROCAP,
            )
        ),
        dna_model_pb2.VariantScorer(
            center_mask=dna_model_pb2.CenterMaskScorer(
                width=None,
                aggregation_type=(
                    dna_model_pb2.AggregationType.AGGREGATION_TYPE_DIFF_LOG2_SUM
                ),
                requested_output=dna_model_pb2.OUTPUT_TYPE_ATAC,
            )
        ),
        dna_model_pb2.VariantScorer(
            gene_mask=dna_model_pb2.GeneMaskLFCScorer(
                requested_output=dna_model_pb2.OUTPUT_TYPE_RNA_SEQ,
            )
        ),
        dna_model_pb2.VariantScorer(
            gene_mask_splicing=dna_model_pb2.GeneMaskSplicingScorer(
                width=10_001,
                requested_output=dna_model_pb2.OUTPUT_TYPE_SPLICE_SITE_USAGE,
            )
        ),
        dna_model_pb2.VariantScorer(
            pa_qtl=dna_model_pb2.PolyadenylationScorer()
        ),
        dna_model_pb2.VariantScorer(
            splice_junction=dna_model_pb2.SpliceJunctionScorer()
        ),
    ]
    scorers = [
        variant_scorers.CenterMaskScorer(
            width=10_001,
            aggregation_type=(variant_scorers.AggregationType.DIFF_MEAN),
            requested_output=dna_output.OutputType.RNA_SEQ,
        ),
        variant_scorers.CenterMaskScorer(
            width=10_001,
            aggregation_type=(variant_scorers.AggregationType.ACTIVE_MEAN),
            requested_output=dna_output.OutputType.PROCAP,
        ),
        variant_scorers.CenterMaskScorer(
            width=None,
            aggregation_type=(variant_scorers.AggregationType.DIFF_LOG2_SUM),
            requested_output=dna_output.OutputType.ATAC,
        ),
        variant_scorers.GeneMaskLFCScorer(
            requested_output=dna_output.OutputType.RNA_SEQ,
        ),
        variant_scorers.GeneMaskSplicingScorer(
            width=10_001,
            requested_output=dna_output.OutputType.SPLICE_SITE_USAGE,
        ),
        variant_scorers.PolyadenylationScorer(),
        variant_scorers.SpliceJunctionScorer(),
    ]
    for variant_scorer, expected_proto in zip(scorers, expected):
      self.assertEqual(variant_scorer.to_proto(), expected_proto)

  def test_unsupported_output_type_raises_error(self):
    with self.assertRaisesRegex(ValueError, 'Unsupported requested output'):
      _ = variant_scorers.CenterMaskScorer(
          width=10_001,
          aggregation_type=variant_scorers.AggregationType.DIFF_MEAN,
          requested_output=dna_output.OutputType.CONTACT_MAPS,
      )

  def test_unsupported_width_raises_error(self):
    with self.assertRaisesRegex(ValueError, 'Unsupported width'):
      _ = variant_scorers.CenterMaskScorer(
          width=1,
          aggregation_type=variant_scorers.AggregationType.DIFF_MEAN,
          requested_output=dna_output.OutputType.RNA_SEQ,
      )


class TestTidyScores(absltest.TestCase):
  """Test tidying scores."""

  def setUp(self):
    """Set up test data."""
    super().setUp()

    # Possible scorers.
    self.gene_scorer = variant_scorers.GeneMaskLFCScorer(
        requested_output=dna_output.OutputType.RNA_SEQ,
    )
    self.center_mask_scorer = variant_scorers.CenterMaskScorer(
        width=10_001,
        aggregation_type=variant_scorers.AggregationType.DIFF_MEAN,
        requested_output=dna_output.OutputType.DNASE,
    )
    self.paqtl_scorer = variant_scorers.PolyadenylationScorer()
    self.junction_scorer = variant_scorers.SpliceJunctionScorer()
    self.splicing_scorer_ssu = variant_scorers.GeneMaskSplicingScorer(
        dna_output.OutputType.SPLICE_SITE_USAGE, width=101
    )
    self.splicing_scorer_ss = variant_scorers.GeneMaskSplicingScorer(
        dna_output.OutputType.SPLICE_SITES, width=101
    )
    self.gene_interval_scorer = interval_scorers.GeneMaskScorer(
        requested_output=dna_output.OutputType.RNA_SEQ,
        width=10_001,
        aggregation_type=interval_scorers.IntervalAggregationType.MEAN,
    )

    # Gene-centric scoring output.
    self.adata_gene_centric = anndata.AnnData(
        X=np.array([[1, 2], [3, 4], [5, 6]]),
        # 3 genes in vicinity of the variant.
        obs=pd.DataFrame({
            'gene_id': [
                'ENSG00000100342',
                'ENSG00000123456',
                'ENSG00000123457',
            ],
            'gene_name': ['GENE1', 'GENE2', 'GENE3'],
            'gene_type': ['protein_coding', 'lncRNA', 'protein_coding'],
            'strand': ['+', '-', '-'],
        }),
        # Two RNA_SEQ tracks.
        var=pd.DataFrame({
            'name': [
                'UBERON:0002107 total RNA-seq',
                'UBERON:0006566 gtex Heart_Left_Ventricle polyA plus RNA-seq',
            ],
            'strand': ['+', '-'],
            'Assay title': ['total RNA-seq', 'polyA plus RNA-seq'],
            'ontology_curie': ['UBERON:0002107', 'UBERON:0006566'],
            'biosample_name': ['liver', 'left ventricle myocardium'],
            'biosample_type': ['tissue', 'tissue'],
            'gtex_tissue': ['Liver', 'Heart_Left_Ventricle'],
        }),
        uns={
            'variant': 'chr1:100:A>C',
            'interval': 'chr1:90-110',
            'variant_scorer': self.gene_scorer,
        },
        layers={
            'quantiles': np.array([[1, 0], [-1, 0], [1, -1]]),
        },
    )

    # Variant-centric scoring output.
    self.adata_variant_centric = anndata.AnnData(
        # 4 DNASE tracks.
        X=np.array([[5, 6, 7, 8]]),
        obs=pd.DataFrame(index=[0]),
        var=pd.DataFrame({
            'name': [
                'UBERON:0001 dnase',
                'UBERON:0002 dnase',
                'UBERON:0003 dnase',
                'UBERON:0004 dnase',
            ],
            'strand': ['-', '+', '-', '.'],
            'Assay title': ['DNase-seq'] * 4,
            'ontology_curie': [
                'UBERON:0001',
                'UBERON:0002',
                'UBERON:0003',
                'UBERON:0004',
            ],
            'biosample_name': [
                'T-cell',
                'endodermal cell',
                'natural killer cell',
                'CD4-positive, alpha-beta memory T cell',
            ],
            'biosample_type': [
                'primary_cell',
                'in_vitro_differentiated_cells',
                'primary_cell',
                'primary_cell',
            ],
        }),
        uns={
            'variant': 'chr2:200:G>T',
            'interval': 'chr2:190-210',
            'variant_scorer': self.center_mask_scorer,
        },
    )
    # PolyadenylationScorer RNA_SEQ scoring output. This output has the same
    # format as a gene-centric scoring output so it is not really a special
    # case.
    self.adata_paqtl = anndata.AnnData(
        X=np.array([
            [
                5,
                6,
                7,
            ],
            [8, 9, 10],
        ]),
        obs=pd.DataFrame({
            'gene_id': ['ENSG00000100342', 'ENSG00000123456'],
            'gene_name': ['GENE1', 'GENE2'],
            'gene_type': ['protein_coding', 'lncRNA'],
            'strand': ['+', '-'],
        }),
        var=pd.DataFrame({
            'name': [
                'CL:0000047 polyA plus RNA-seq',
                'CL:0000047 polyA plus RNA-seq',
                'UBERON:0018115 polyA plus RNA-seq',
            ],
            'strand': ['+', '-', '.'],
            'Assay title': [
                'polyA plus RNA-seq',
                'polyA plus RNA-seq',
                'polyA plus RNA-seq',
            ],
            'ontology_curie': [
                'CL:0000047',
                'CL:0000047',
                'UBERON:0018115',
            ],
            'biosample_name': [
                'neuronal stem cell',
                'neuronal stem cell',
                'left renal pelvis',
            ],
            'biosample_type': [
                'in_vitro_differentiated_cells',
                'in_vitro_differentiated_cells',
                'tissue',
            ],
            'gtex_tissue': ['', '', ''],
        }),
        uns={
            'variant': 'chr3:300:A>G',
            'interval': 'chr3:290-310',
            'variant_scorer': self.paqtl_scorer,
        },
    )
    # SpliceJunctionScorer SPLICE_JUNCTIONS scoring output.
    self.adata_junction = anndata.AnnData(
        X=np.array([
            [5, 6, 7, 8],
            [9, 10, 11, 12],
        ]),
        obs=pd.DataFrame({
            'gene_id': ['ENSG00000100342', 'ENSG00000123456'],
            'gene_name': ['GENE1', 'GENE2'],
            'gene_type': ['protein_coding', 'lncRNA'],
            'strand': ['+', '-'],
            'junction_Start': [0, 1],
            'junction_End': [1, 2],
        }),
        var=pd.DataFrame({
            'name': [
                'delta_psi5:Brain_Cerebellum',
                'delta_psi3:Brain_Cerebellum',
                'delta_counts:Brain_Cerebellum',
                'ref_counts:Brain_Cerebellum',
            ],
            'strand': ['.', '.', '.', '.'],
            'gtex_tissue': [
                'Brain_Cerebellum',
                'Brain_Cerebellum',
                'Brain_Cerebellum',
                'Brain_Cerebellum',
            ],
        }),
        uns={
            'variant': 'chr3:300:A>G',
            'interval': 'chr3:290-310',
            'variant_scorer': self.junction_scorer,
        },
    )
    # GeneMaskSplicingScorer SPLICE_SITE_USAGE scoring output.
    self.adata_splicing_ssu = anndata.AnnData(
        X=np.array([[1, 2], [3, 4]]),
        obs=pd.DataFrame({
            'gene_id': ['ENSG00000100342', 'ENSG00000123456'],
            'gene_name': ['GENE1', 'GENE2'],
            'gene_type': ['protein_coding', 'lncRNA'],
            'strand': ['+', '-'],
        }),
        var=pd.DataFrame({
            'name': [
                'usage_Brain_Cerebellum',
                'usage_Brain_Cerebellum',
            ],
            'strand': ['+', '-'],
            'ontology_curie': ['UBERON:0002037', 'UBERON:0002037'],
            'biosample_name': ['cerebellum', 'cerebellum'],
            'gtex_tissue': [
                'Brain_Cerebellum',
                'Brain_Cerebellum',
            ],
        }),
        uns={
            'variant': 'chr3:300:A>G',
            'interval': 'chr3:290-310',
            'variant_scorer': self.splicing_scorer_ssu,
        },
    )
    # GeneMaskSplicingScorer SPLICE_SITES scoring output.
    self.adata_splicing_ss = anndata.AnnData(
        X=np.array([[1, 2, 3, 4, 5]]),
        obs=pd.DataFrame({
            'gene_id': ['ENSG00000100342'],
            'gene_name': ['GENE1'],
            'gene_type': ['protein_coding'],
            'strand': ['+'],
        }),
        var=pd.DataFrame({
            # There are only ever these 4 tracks for this scorer.
            'name': [
                'donor',
                'acceptor',
                'donor',
                'acceptor',
                'other',
            ],
            'strand': ['+', '+', '-', '-', '.'],
        }),
        uns={
            'variant': 'chr3:300:A>G',
            'interval': 'chr3:290-310',
            'variant_scorer': self.splicing_scorer_ss,
        },
    )

    self.extended_metadata_columns = set([
        'Assay title',
        'ontology_curie',
        'biosample_name',
        'biosample_type',
        'gtex_tissue',
        'transcription_factor',
        'histone_mark',
    ])
    # Interval scoring output.
    self.adata_interval_centric = anndata.AnnData(
        X=np.array([[1, 2], [3, 4], [5, 6]]),
        # 3 genes in vicinity of the interval.
        obs=pd.DataFrame({
            'gene_id': [
                'ENSG00000100342',
                'ENSG00000123456',
                'ENSG00000123457',
            ],
            'gene_name': ['GENE1', 'GENE2', 'GENE3'],
            'gene_type': ['protein_coding', 'lncRNA', 'protein_coding'],
            'strand': ['+', '-', '-'],
        }),
        # Two RNA_SEQ tracks.
        var=pd.DataFrame({
            'name': [
                'UBERON:0002107 total RNA-seq',
                'UBERON:0006566 gtex Heart_Left_Ventricle polyA plus RNA-seq',
            ],
            'strand': ['+', '-'],
            'Assay title': ['total RNA-seq', 'polyA plus RNA-seq'],
            'ontology_curie': ['UBERON:0002107', 'UBERON:0006566'],
            'biosample_name': ['liver', 'left ventricle myocardium'],
            'biosample_type': ['tissue', 'tissue'],
            'gtex_tissue': ['Liver', 'Heart_Left_Ventricle'],
        }),
        uns={
            'interval': 'chr1:90-110',
            'interval_scorer': self.gene_interval_scorer,
        },
    )

  def test_tidy_anndata_gene_centric(self):
    """Test tidying a single gene centric AnnData object."""
    df = variant_scorers.tidy_anndata(
        self.adata_gene_centric, match_gene_strand=False
    )
    self.assertLen(df, 6)  # 3 genes by 2 tracks.

    df = variant_scorers.tidy_anndata(self.adata_gene_centric)
    self.assertLen(df, 3)  # Removes mismatched strands.
    expected = pd.DataFrame({
        'variant_id': ['chr1:100:A>C', 'chr1:100:A>C', 'chr1:100:A>C'],
        'scored_interval': ['chr1:90-110', 'chr1:90-110', 'chr1:90-110'],
        'gene_id': ['ENSG00000100342', 'ENSG00000123456', 'ENSG00000123457'],
        'gene_name': ['GENE1', 'GENE2', 'GENE3'],
        'gene_type': ['protein_coding', 'lncRNA', 'protein_coding'],
        'gene_strand': ['+', '-', '-'],
        'junction_Start': [None] * 3,
        'junction_End': [None] * 3,
        'output_type': ['RNA_SEQ', 'RNA_SEQ', 'RNA_SEQ'],
        'variant_scorer': [str(self.gene_scorer)] * 3,
        'track_name': [
            'UBERON:0002107 total RNA-seq',
            'UBERON:0006566 gtex Heart_Left_Ventricle polyA plus RNA-seq',
            'UBERON:0006566 gtex Heart_Left_Ventricle polyA plus RNA-seq',
        ],
        'track_strand': ['+', '-', '-'],
        'Assay title': [
            'total RNA-seq',
            'polyA plus RNA-seq',
            'polyA plus RNA-seq',
        ],
        'ontology_curie': [
            'UBERON:0002107',
            'UBERON:0006566',
            'UBERON:0006566',
        ],
        'biosample_name': [
            'liver',
            'left ventricle myocardium',
            'left ventricle myocardium',
        ],
        'biosample_type': ['tissue', 'tissue', 'tissue'],
        'gtex_tissue': [
            'Liver',
            'Heart_Left_Ventricle',
            'Heart_Left_Ventricle',
        ],
        'raw_score': [1, 4, 6],
        'quantile_score': [1, 0, -1],
    })
    pd.testing.assert_frame_equal(df, expected)
    # Test without extended metadata.
    df = variant_scorers.tidy_anndata(
        self.adata_gene_centric, include_extended_metadata=False
    )
    pd.testing.assert_frame_equal(
        df,
        expected[
            [c for c in df.columns if c not in self.extended_metadata_columns]
        ],
    )

  def test_tidy_anndata_variant_centric(self):
    """Test tidying a single center mask AnnData object."""
    df = variant_scorers.tidy_anndata(self.adata_variant_centric)
    self.assertLen(df, 4)  # 4 tracks.
    # Strand filtering has no effect on variant-centric scoring.
    df = variant_scorers.tidy_anndata(
        self.adata_variant_centric, match_gene_strand=False
    )
    self.assertLen(df, 4)

    expected = pd.DataFrame({
        'variant_id': [
            'chr2:200:G>T',
            'chr2:200:G>T',
            'chr2:200:G>T',
            'chr2:200:G>T',
        ],
        'scored_interval': [
            'chr2:190-210',
            'chr2:190-210',
            'chr2:190-210',
            'chr2:190-210',
        ],
        'gene_id': [None, None, None, None],
        'gene_name': [None, None, None, None],
        'gene_type': [None, None, None, None],
        'gene_strand': [None, None, None, None],
        'junction_Start': [None] * 4,
        'junction_End': [None] * 4,
        'output_type': ['DNASE', 'DNASE', 'DNASE', 'DNASE'],
        'variant_scorer': [str(self.center_mask_scorer)] * 4,
        'track_name': [
            'UBERON:0001 dnase',
            'UBERON:0002 dnase',
            'UBERON:0003 dnase',
            'UBERON:0004 dnase',
        ],
        'track_strand': ['-', '+', '-', '.'],
        'Assay title': ['DNase-seq', 'DNase-seq', 'DNase-seq', 'DNase-seq'],
        'ontology_curie': [
            'UBERON:0001',
            'UBERON:0002',
            'UBERON:0003',
            'UBERON:0004',
        ],
        'biosample_name': [
            'T-cell',
            'endodermal cell',
            'natural killer cell',
            'CD4-positive, alpha-beta memory T cell',
        ],
        'biosample_type': [
            'primary_cell',
            'in_vitro_differentiated_cells',
            'primary_cell',
            'primary_cell',
        ],
        'raw_score': [5, 6, 7, 8],
    })
    pd.testing.assert_frame_equal(df, expected)
    # Test without extended metadata.
    df = variant_scorers.tidy_anndata(
        self.adata_variant_centric, include_extended_metadata=False
    )
    pd.testing.assert_frame_equal(
        df,
        expected[
            [c for c in df.columns if c not in self.extended_metadata_columns]
        ],
    )

  def test_tidy_anndata_paqtl(self):
    """Test tidying a single paQTL scoring output AnnData object."""
    df = variant_scorers.tidy_anndata(self.adata_paqtl)

    expected = pd.DataFrame({
        'variant_id': [
            'chr3:300:A>G',
            'chr3:300:A>G',
            'chr3:300:A>G',
            'chr3:300:A>G',
        ],
        'scored_interval': [
            'chr3:290-310',
            'chr3:290-310',
            'chr3:290-310',
            'chr3:290-310',
        ],
        'gene_id': [
            'ENSG00000100342',
            'ENSG00000123456',
            'ENSG00000100342',
            'ENSG00000123456',
        ],
        'gene_name': ['GENE1', 'GENE2', 'GENE1', 'GENE2'],
        'gene_type': ['protein_coding', 'lncRNA', 'protein_coding', 'lncRNA'],
        'gene_strand': ['+', '-', '+', '-'],
        'junction_Start': [None] * 4,
        'junction_End': [None] * 4,
        'output_type': ['RNA_SEQ', 'RNA_SEQ', 'RNA_SEQ', 'RNA_SEQ'],
        'variant_scorer': [str(self.paqtl_scorer)] * 4,
        'track_name': [
            'CL:0000047 polyA plus RNA-seq',
            'CL:0000047 polyA plus RNA-seq',
            'UBERON:0018115 polyA plus RNA-seq',
            'UBERON:0018115 polyA plus RNA-seq',
        ],
        'track_strand': ['+', '-', '.', '.'],
        'Assay title': [
            'polyA plus RNA-seq',
            'polyA plus RNA-seq',
            'polyA plus RNA-seq',
            'polyA plus RNA-seq',
        ],
        'ontology_curie': [
            'CL:0000047',
            'CL:0000047',
            'UBERON:0018115',
            'UBERON:0018115',
        ],
        'biosample_name': [
            'neuronal stem cell',
            'neuronal stem cell',
            'left renal pelvis',
            'left renal pelvis',
        ],
        'biosample_type': [
            'in_vitro_differentiated_cells',
            'in_vitro_differentiated_cells',
            'tissue',
            'tissue',
        ],
        'gtex_tissue': ['', '', '', ''],
        'raw_score': [5, 9, 7, 10],
    })
    pd.testing.assert_frame_equal(df, expected)
    # Test without extended metadata.
    df = variant_scorers.tidy_anndata(
        self.adata_paqtl, include_extended_metadata=False
    )
    pd.testing.assert_frame_equal(
        df,
        expected[
            [c for c in df.columns if c not in self.extended_metadata_columns]
        ],
    )

  def test_tidy_anndata_interval_centric(self):
    """Test tidying a single gene centric interval scoring AnnData object."""
    df = variant_scorers.tidy_anndata(
        self.adata_interval_centric, match_gene_strand=False
    )
    self.assertLen(df, 6)  # 3 genes by 2 tracks.

    df = variant_scorers.tidy_anndata(self.adata_interval_centric)
    self.assertLen(df, 3)  # Removes mismatched strands.
    expected = pd.DataFrame({
        'scored_interval': ['chr1:90-110', 'chr1:90-110', 'chr1:90-110'],
        'gene_id': ['ENSG00000100342', 'ENSG00000123456', 'ENSG00000123457'],
        'gene_name': ['GENE1', 'GENE2', 'GENE3'],
        'gene_type': ['protein_coding', 'lncRNA', 'protein_coding'],
        'gene_strand': ['+', '-', '-'],
        'junction_Start': [None] * 3,
        'junction_End': [None] * 3,
        'output_type': ['RNA_SEQ', 'RNA_SEQ', 'RNA_SEQ'],
        'interval_scorer': [str(self.gene_interval_scorer)] * 3,
        'track_name': [
            'UBERON:0002107 total RNA-seq',
            'UBERON:0006566 gtex Heart_Left_Ventricle polyA plus RNA-seq',
            'UBERON:0006566 gtex Heart_Left_Ventricle polyA plus RNA-seq',
        ],
        'track_strand': ['+', '-', '-'],
        'Assay title': [
            'total RNA-seq',
            'polyA plus RNA-seq',
            'polyA plus RNA-seq',
        ],
        'ontology_curie': [
            'UBERON:0002107',
            'UBERON:0006566',
            'UBERON:0006566',
        ],
        'biosample_name': [
            'liver',
            'left ventricle myocardium',
            'left ventricle myocardium',
        ],
        'biosample_type': ['tissue', 'tissue', 'tissue'],
        'gtex_tissue': [
            'Liver',
            'Heart_Left_Ventricle',
            'Heart_Left_Ventricle',
        ],
        'raw_score': [1, 4, 6],
    })
    pd.testing.assert_frame_equal(df, expected)
    # Test without extended metadata.
    df = variant_scorers.tidy_anndata(
        self.adata_interval_centric, include_extended_metadata=False
    )
    pd.testing.assert_frame_equal(
        df,
        expected[
            [c for c in df.columns if c not in self.extended_metadata_columns]
        ],
    )

  def test_tidy_anndata_junction(self):
    """Test tidying a single splice junction scoring output AnnData object."""
    df = variant_scorers.tidy_anndata(self.adata_junction)
    expected = pd.DataFrame({
        'variant_id': [
            'chr3:300:A>G',
            'chr3:300:A>G',
            'chr3:300:A>G',
            'chr3:300:A>G',
            'chr3:300:A>G',
            'chr3:300:A>G',
            'chr3:300:A>G',
            'chr3:300:A>G',
        ],
        'scored_interval': [
            'chr3:290-310',
            'chr3:290-310',
            'chr3:290-310',
            'chr3:290-310',
            'chr3:290-310',
            'chr3:290-310',
            'chr3:290-310',
            'chr3:290-310',
        ],
        'gene_id': [
            'ENSG00000100342',
            'ENSG00000123456',
            'ENSG00000100342',
            'ENSG00000123456',
            'ENSG00000100342',
            'ENSG00000123456',
            'ENSG00000100342',
            'ENSG00000123456',
        ],
        'gene_name': [
            'GENE1',
            'GENE2',
            'GENE1',
            'GENE2',
            'GENE1',
            'GENE2',
            'GENE1',
            'GENE2',
        ],
        'gene_type': [
            'protein_coding',
            'lncRNA',
            'protein_coding',
            'lncRNA',
            'protein_coding',
            'lncRNA',
            'protein_coding',
            'lncRNA',
        ],
        'gene_strand': ['+', '-', '+', '-', '+', '-', '+', '-'],
        'junction_Start': [0, 1, 0, 1, 0, 1, 0, 1],
        'junction_End': [1, 2, 1, 2, 1, 2, 1, 2],
        'output_type': [
            'SPLICE_JUNCTIONS',
            'SPLICE_JUNCTIONS',
            'SPLICE_JUNCTIONS',
            'SPLICE_JUNCTIONS',
            'SPLICE_JUNCTIONS',
            'SPLICE_JUNCTIONS',
            'SPLICE_JUNCTIONS',
            'SPLICE_JUNCTIONS',
        ],
        'variant_scorer': [str(self.junction_scorer)] * 8,
        'track_name': [
            'delta_psi5:Brain_Cerebellum',
            'delta_psi5:Brain_Cerebellum',
            'delta_psi3:Brain_Cerebellum',
            'delta_psi3:Brain_Cerebellum',
            'delta_counts:Brain_Cerebellum',
            'delta_counts:Brain_Cerebellum',
            'ref_counts:Brain_Cerebellum',
            'ref_counts:Brain_Cerebellum',
        ],
        'track_strand': ['.', '.', '.', '.', '.', '.', '.', '.'],
        'gtex_tissue': ['Brain_Cerebellum'] * 8,
        'raw_score': [5, 9, 6, 10, 7, 11, 8, 12],
    })
    pd.testing.assert_frame_equal(df, expected)
    # Test without extended metadata.
    df = variant_scorers.tidy_anndata(
        self.adata_junction, include_extended_metadata=False
    )
    pd.testing.assert_frame_equal(
        df,
        expected[
            [c for c in df.columns if c not in self.extended_metadata_columns]
        ],
    )

  def test_tidy_anndata_splicing_ssu(self):
    """Test tidying a single splicing ssu scoring output AnnData object."""
    df = variant_scorers.tidy_anndata(self.adata_splicing_ssu)
    expected = pd.DataFrame({
        'variant_id': [
            'chr3:300:A>G',
            'chr3:300:A>G',
        ],
        'scored_interval': [
            'chr3:290-310',
            'chr3:290-310',
        ],
        'gene_id': [
            'ENSG00000100342',
            'ENSG00000123456',
        ],
        'gene_name': [
            'GENE1',
            'GENE2',
        ],
        'gene_type': ['protein_coding', 'lncRNA'],
        'gene_strand': ['+', '-'],
        'junction_Start': [None] * 2,
        'junction_End': [None] * 2,
        'output_type': [
            'SPLICE_SITE_USAGE',
            'SPLICE_SITE_USAGE',
        ],
        'variant_scorer': [str(self.splicing_scorer_ssu)] * 2,
        'track_name': [
            'usage_Brain_Cerebellum',
            'usage_Brain_Cerebellum',
        ],
        'track_strand': [
            '+',
            '-',
        ],
        'ontology_curie': ['UBERON:0002037', 'UBERON:0002037'],
        'biosample_name': ['cerebellum', 'cerebellum'],
        'gtex_tissue': [
            'Brain_Cerebellum',
            'Brain_Cerebellum',
        ],
        'raw_score': [1, 4],
    })
    pd.testing.assert_frame_equal(df, expected)

  def test_tidy_anndata_splicing_sites(self):
    """Test tidying a single splicing sites scoring output AnnData object."""
    df = variant_scorers.tidy_anndata(self.adata_splicing_ss)
    expected = pd.DataFrame({
        'variant_id': ['chr3:300:A>G', 'chr3:300:A>G', 'chr3:300:A>G'],
        'scored_interval': ['chr3:290-310', 'chr3:290-310', 'chr3:290-310'],
        'gene_id': ['ENSG00000100342', 'ENSG00000100342', 'ENSG00000100342'],
        'gene_name': ['GENE1', 'GENE1', 'GENE1'],
        'gene_type': ['protein_coding', 'protein_coding', 'protein_coding'],
        'gene_strand': ['+', '+', '+'],
        'junction_Start': [None] * 3,
        'junction_End': [None] * 3,
        'output_type': ['SPLICE_SITES', 'SPLICE_SITES', 'SPLICE_SITES'],
        'variant_scorer': [str(self.splicing_scorer_ss)] * 3,
        'track_name': ['donor', 'acceptor', 'other'],
        'track_strand': ['+', '+', '.'],
        'raw_score': [1, 2, 5],
    })
    pd.testing.assert_frame_equal(df, expected)
    # Test without extended metadata.
    df = variant_scorers.tidy_anndata(
        self.adata_splicing_ss, include_extended_metadata=False
    )
    pd.testing.assert_frame_equal(
        df,
        expected[
            [c for c in df.columns if c not in self.extended_metadata_columns]
        ],
    )

  def test_tidy_lists_of_anndata(self):
    # List of AnnDatas.
    anndata_list = [self.adata_gene_centric, self.adata_variant_centric]
    df_list = variant_scorers.tidy_scores(anndata_list)
    self.assertLen(df_list, 7)

    # List of list of AnnDatas.
    anndata_list = [[self.adata_gene_centric], [self.adata_variant_centric]]
    df_nested_list = variant_scorers.tidy_scores(anndata_list)
    self.assertLen(df_nested_list, 7)
    # The exact nesting pattern doesn't matter, the result is equivalent.
    pd.testing.assert_frame_equal(df_list, df_nested_list)

    # Longer list of list of AnnDatas.
    anndata_list = [
        [self.adata_gene_centric, self.adata_variant_centric],
        [self.adata_paqtl, self.adata_junction, self.adata_interval_centric],
    ]
    df_longer_nested_list = variant_scorers.tidy_scores(anndata_list)
    self.assertLen(df_longer_nested_list, 22)

    # Scores are empty.
    anndata_list = [anndata.AnnData(X=np.ones((0, 3)))]
    df_empty_list = variant_scorers.tidy_scores(anndata_list)
    self.assertIsNone(df_empty_list)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest
from absl.testing import parameterized
from alphagenome.data import genome
from alphagenome.data import track_data
import jaxtyping
import numpy as np
import pandas as pd


class TrackDataInitTest(parameterized.TestCase):

  @parameterized.product(copy=[True, False], include_ontologies=[True, False])
  def test_valid_init_properties(self, copy: bool, include_ontologies: bool):
    values = np.arange(15).reshape((5, 3)).astype(np.int32)
    metadata = pd.DataFrame({
        'name': ['first', 'first', 'second'],
        'strand': ['+', '-', '.'],
        'other': ['foo', 'foo', 'bar'],
    })
    if include_ontologies:
      metadata['ontology_curie'] = [
          'UBERON:0000005',
          'UBERON:0000007',
          None,
      ]
    tdata = track_data.TrackData(values, metadata)
    if copy:
      tdata = tdata.copy()
    self.assertListEqual(tdata.positional_axes, [0])
    self.assertEqual(tdata.num_tracks, 3)
    self.assertEqual(tdata.width, 5)
    self.assertEqual(tdata.resolution, 1)
    self.assertListEqual(list(tdata.names), ['first', 'first', 'second'])
    self.assertListEqual(list(tdata.strands), ['+', '-', '.'])
    if include_ontologies:
      self.assertEqual(
          [
              o.ontology_curie if o is not None else None
              for o in tdata.ontology_terms
          ],
          metadata['ontology_curie'].tolist(),
      )
    else:
      self.assertIsNone(tdata.ontology_terms)

  @parameterized.parameters([
      dict(num_positional=0),
      dict(num_positional=1),
      dict(num_positional=2),
  ])
  def test_valid_positional_axis(self, num_positional: int):
    values = (
        np.arange(2**num_positional * 3)
        .reshape([2] * num_positional + [3])
        .astype(np.int32)
    )
    metadata = pd.DataFrame({
        'name': ['first', 'first', 'second'],
        'strand': ['+', '-', '.'],
        'other': ['foo', 'foo', 'bar'],
    })
    tdata = track_data.TrackData(values, metadata)
    self.assertLen(tdata.positional_axes, num_positional)

  def test_invalid_init(self):
    values = np.arange(12).reshape((2, 2, 3)).astype(np.int32)
    metadata = pd.DataFrame({
        'name': ['first', 'first', 'second'],
        'strand': ['+', '-', '.'],
        'other': ['foo', 'foo', 'bar'],
    })
    with self.assertRaises(ValueError):
      # Missing strand
      track_data.TrackData(values, metadata[['name', 'other']])
    with self.assertRaises(ValueError):
      # Missing name
      track_data.TrackData(values, metadata[['strand', 'other']])
    with self.assertRaises(ValueError):
      # Not matching number of tracks
      track_data.TrackData(values[:, :, :2], metadata)
    with self.assertRaises(ValueError):
      # Not matching positional axis
      track_data.TrackData(values[:1], metadata)
    with self.assertRaises(ValueError):
      # Not matching interval width
      track_data.TrackData(
          values, metadata, interval=genome.Interval('chr1', 0, 3)
      )

  def test_invalid_value_types_raises(self):
    values = np.array(['foo', 'bar', 'baz'], dtype=str)
    metadata = pd.DataFrame({
        'name': ['first', 'first', 'second'],
        'strand': ['+', '-', '.'],
        'other': ['foo', 'foo', 'bar'],
    })
    with self.assertRaises(jaxtyping.TypeCheckError):
      track_data.TrackData(values, metadata)


class TrackDataSlicingTest(parameterized.TestCase):

  @parameterized.parameters([
      dict(num_positional=0),
      dict(num_positional=1),
      dict(num_positional=2),
  ])
  def test_slice_by_positions(self, num_positional: int):
    # Values are [0,1,2] if num_positional=0, [[0,1,2],[3,4,5]] if
    # num_positional=1, etc.
    values = (
        np.arange(2**num_positional * 3)
        .reshape([2] * num_positional + [3])
        .astype(np.int32)
    )
    metadata = pd.DataFrame({
        'name': ['first', 'first', 'second'],
        'strand': ['+', '-', '.'],
        'other': ['foo', 'foo', 'bar'],
    })
    interval = genome.Interval('chr1', 0, 4)
    tdata = track_data.TrackData(
        values, metadata, resolution=2, interval=interval
    )
    assert tdata.interval is not None
    self.assertIsNotNone(tdata.interval)
    self.assertEqual(tdata.interval.start, 0)
    self.assertEqual(tdata.interval.end, 4)
    # Note that track data with no positional axes has a width of 0 and so
    # cannot be sliced in this way.
    if num_positional:
      # Slicing to the full width (no-op).
      tdata2 = tdata.slice_by_positions(0, 4)
      assert tdata2.interval is not None
      self.assertEqual(tdata2.width, tdata.width)
      # Slicing to half the width.
      tdata2 = tdata.slice_by_positions(0, 2)
      self.assertEqual(tdata2.values.shape, tuple([1] * num_positional + [3]))
      self.assertEqual(tdata2.interval.start, 0)
      self.assertEqual(tdata2.interval.end, 2)
      self.assertEqual(tdata2.width, tdata2.interval.width)
      self.assertEqual(tdata2.width, 2)
      # Fails because (end - start) is not evenly divisible by resolution.
      with self.assertRaises(ValueError):
        tdata.slice_by_positions(0, 3)
      with self.assertRaises(ValueError):
        tdata.slice_by_positions(0, 1)
    # Fails because there are no positional axes to slice into.
    else:
      self.assertEqual(tdata.width, 0)
      with self.assertRaises(ValueError):
        tdata.slice_by_positions(0, 4)

  @parameterized.parameters([
      dict(num_positional=0),
      dict(num_positional=1),
      dict(num_positional=2),
  ])
  def test_slice_by_interval(self, num_positional: int):
    # Values are [0,1,2] if num_positional=0, [[0,1,2],[3,4,5]] if
    # num_positional=1, etc.
    values = (
        np.arange(2**num_positional * 3)
        .reshape([2] * num_positional + [3])
        .astype(np.int32)
    )
    metadata = pd.DataFrame({
        'name': ['first', 'first', 'second'],
        'strand': ['+', '-', '.'],
        'other': ['foo', 'foo', 'bar'],
    })
    interval = genome.Interval('chr1', 1, 5)
    tdata = track_data.TrackData(
        values, metadata, resolution=2, interval=interval
    )
    assert tdata.interval is not None

    if num_positional:
      # Slicing to the full width (no-op).
      tdata2 = tdata.slice_by_interval(interval)
      assert tdata2.interval is not None
      self.assertEqual(tdata2.width, tdata.width)
      # Slicing to half the width.
      tdata2 = tdata.slice_by_interval(genome.Interval('chr1', 1, 3))
      self.assertEqual(tdata2.values.shape, tuple([1] * num_positional + [3]))
      self.assertEqual(tdata2.interval.start, 1)
      self.assertEqual(tdata2.interval.end, 3)
      self.assertEqual(tdata2.width, tdata2.interval.width)
      self.assertEqual(tdata2.width, 2)
      # Fails because (end - start) is not evenly divisible by resolution.
      with self.assertRaises(ValueError):
        tdata.slice_by_interval(genome.Interval('chr1', 1, 4))
      with self.assertRaises(ValueError):
        tdata.slice_by_interval(genome.Interval('chr1', 1, 2))

      # Same would work if we set match_resolution=True.
      tdata2 = tdata.slice_by_interval(
          genome.Interval('chr1', 1, 4), match_resolution=True
      )
      self.assertEqual(tdata2.interval.start, 1)
      self.assertEqual(tdata2.interval.end, 5)
      tdata2 = tdata.slice_by_interval(
          genome.Interval('chr1', 2, 5), match_resolution=True
      )
      self.assertEqual(tdata2.interval.start, 1)
      self.assertEqual(tdata2.interval.end, 5)
      tdata2 = tdata.slice_by_interval(
          genome.Interval('chr1', 3, 5), match_resolution=True
      )
      self.assertEqual(tdata2.interval.start, 3)
      self.assertEqual(tdata2.interval.end, 5)
    else:
      # Fails because there are no positional axes to slice into.
      self.assertEqual(tdata.width, 0)
      with self.assertRaises(ValueError):
        tdata.slice_by_interval(interval)

  @parameterized.parameters([
      dict(num_positional=0),
      dict(num_positional=1),
      dict(num_positional=2),
  ])
  def test_pad(self, num_positional: int):
    values = (
        np.arange(2**num_positional * 3)
        .reshape([2] * num_positional + [3])
        .astype(np.int32)
    )
    metadata = pd.DataFrame({
        'name': ['first', 'first', 'second'],
        'strand': ['+', '-', '.'],
        'other': ['foo', 'foo', 'bar'],
    })
    tdata = track_data.TrackData(values, metadata, resolution=2)
    if num_positional:
      self.assertEqual(tdata.width, 4)
    tdata2 = tdata.pad(0, 0)
    self.assertEqual(tdata2.width, tdata.width)
    tdata2 = tdata.pad(2, 0)
    self.assertEqual(tdata2.values.shape, tuple([3] * num_positional + [3]))
    if num_positional:
      self.assertEqual(tdata2.width, 6)
    with self.assertRaises(ValueError):
      tdata.pad(1, 1)

  @parameterized.parameters([
      dict(num_positional=0),
      dict(num_positional=1),
      dict(num_positional=2),
  ])
  def test_resize(self, num_positional: int):
    values = (
        np.arange(2**num_positional * 3)
        .reshape([2] * num_positional + [3])
        .astype(np.int32)
    )
    metadata = pd.DataFrame({
        'name': ['first', 'first', 'second'],
        'strand': ['+', '-', '.'],
        'other': ['foo', 'foo', 'bar'],
    })
    tdata = track_data.TrackData(
        values,
        metadata,
        resolution=2,
        interval=(genome.Interval('chr1', 0, 4) if num_positional else None),
    )
    if num_positional:
      self.assertEqual(tdata.width, 4)
    tdata2 = tdata.resize(4)
    self.assertEqual(tdata2.width, tdata.width)
    tdata2 = tdata.resize(6)
    self.assertEqual(tdata2.values.shape, tuple([3] * num_positional + [3]))
    if num_positional:
      self.assertEqual(tdata2.width, 6)
    tdata2 = tdata.resize(2)
    self.assertEqual(tdata2.values.shape, tuple([1] * num_positional + [3]))
    if num_positional:
      self.assertEqual(tdata2.width, 2)
    # Pad and crop to reach the same original array.
    tdata2 = tdata.resize(6).resize(4)
    if num_positional:
      self.assertEqual(tdata2.width, 4)
    self.assertTrue(np.all(tdata2.values == tdata.values))
    if num_positional:
      with self.assertRaises(ValueError):
        tdata.resize(5)


class TrackDataUpDownSampleTest(parameterized.TestCase):

  @parameterized.parameters([
      dict(num_positional=0, aggregation_type=track_data.AggregationType.SUM),
      dict(num_positional=1, aggregation_type=track_data.AggregationType.SUM),
      dict(num_positional=2, aggregation_type=track_data.AggregationType.SUM),
      dict(num_positional=0, aggregation_type=track_data.AggregationType.MAX),
      dict(num_positional=1, aggregation_type=track_data.AggregationType.MAX),
      dict(num_positional=2, aggregation_type=track_data.AggregationType.MAX),
  ])
  def test_upsample(
      self, num_positional: int, aggregation_type: track_data.AggregationType
  ):
    values = (
        np.arange(2**num_positional * 3)
        .reshape([2] * num_positional + [3])
        .astype(np.float32)
    )
    metadata = pd.DataFrame({
        'name': ['first', 'first', 'second'],
        'strand': ['+', '-', '.'],
        'other': ['foo', 'foo', 'bar'],
    })
    tdata = track_data.TrackData(values, metadata, resolution=2)
    tdata2 = tdata.upsample(resolution=1, aggregation_type=aggregation_type)
    self.assertEqual(tdata2.width, tdata.width)
    self.assertEqual(tdata2.values.shape, tuple([4] * num_positional + [3]))
    if aggregation_type == track_data.AggregationType.SUM:
      self.assertEqual(tdata2.values.sum(), tdata.values.sum())
    elif aggregation_type == track_data.AggregationType.MAX:
      self.assertEqual(tdata2.values.max(), tdata.values.max())
    with self.assertRaises(ValueError):
      tdata.upsample(resolution=4)

  @parameterized.parameters([
      dict(
          num_positional=0,
          aggregation_type=track_data.AggregationType.SUM,
      ),
      dict(
          num_positional=1,
          aggregation_type=track_data.AggregationType.SUM,
      ),
      dict(
          num_positional=2,
          aggregation_type=track_data.AggregationType.SUM,
      ),
      dict(
          num_positional=0,
          aggregation_type=track_data.AggregationType.MAX,
      ),
      dict(
          num_positional=1,
          aggregation_type=track_data.AggregationType.MAX,
      ),
      dict(
          num_positional=2,
          aggregation_type=track_data.AggregationType.MAX,
      ),
  ])
  def test_downsample(
      self, num_positional: int, aggregation_type: track_data.AggregationType
  ):
    values = (
        np.arange(4**num_positional * 3)
        .reshape([4] * num_positional + [3])
        .astype(np.float32)
    )
    metadata = pd.DataFrame({
        'name': ['first', 'first', 'second'],
        'strand': ['+', '-', '.'],
        'other': ['foo', 'foo', 'bar'],
    })
    tdata = track_data.TrackData(values, metadata, resolution=1)
    tdata2 = tdata.downsample(resolution=2, aggregation_type=aggregation_type)
    self.assertEqual(tdata2.width, tdata.width)
    self.assertEqual(tdata2.values.shape, tuple([2] * num_positional + [3]))
    if aggregation_type == track_data.AggregationType.SUM:
      self.assertEqual(tdata2.values.sum(), tdata.values.sum())
    elif aggregation_type == track_data.AggregationType.MAX:
      self.assertEqual(tdata2.values.max(), tdata.values.max())
    tdata = track_data.TrackData(values, metadata, resolution=2)
    with self.assertRaises(ValueError):
      tdata.downsample(resolution=1)

  @parameterized.parameters([
      dict(
          num_positional=0,
          resolution=1,
          aggregation_type=track_data.AggregationType.SUM,
      ),
      dict(
          num_positional=1,
          resolution=1,
          aggregation_type=track_data.AggregationType.SUM,
      ),
      dict(
          num_positional=2,
          resolution=1,
          aggregation_type=track_data.AggregationType.SUM,
      ),
      dict(
          num_positional=0,
          resolution=4,
          aggregation_type=track_data.AggregationType.SUM,
      ),
      dict(
          num_positional=1,
          resolution=4,
          aggregation_type=track_data.AggregationType.SUM,
      ),
      dict(
          num_positional=2,
          resolution=4,
          aggregation_type=track_data.AggregationType.SUM,
      ),
      dict(
          num_positional=0,
          resolution=1,
          aggregation_type=track_data.AggregationType.MAX,
      ),
      dict(
          num_positional=1,
          resolution=1,
          aggregation_type=track_data.AggregationType.MAX,
      ),
      dict(
          num_positional=2,
          resolution=1,
          aggregation_type=track_data.AggregationType.MAX,
      ),
      dict(
          num_positional=0,
          resolution=4,
          aggregation_type=track_data.AggregationType.MAX,
      ),
      dict(
          num_positional=1,
          resolution=4,
          aggregation_type=track_data.AggregationType.MAX,
      ),
      dict(
          num_positional=2,
          resolution=4,
          aggregation_type=track_data.AggregationType.MAX,
      ),
  ])
  def test_change_resolution(
      self,
      num_positional: int,
      resolution: int,
      aggregation_type: track_data.AggregationType,
  ):
    values = (
        np.arange(2**num_positional * 3)
        .reshape([2] * num_positional + [3])
        .astype(np.float32)
    )
    metadata = pd.DataFrame({
        'name': ['first', 'first', 'second'],
        'strand': ['+', '-', '.'],
        'other': ['foo', 'foo', 'bar'],
    })
    tdata = track_data.TrackData(values, metadata, resolution=2)
    tdata2 = tdata.change_resolution(
        resolution=resolution, aggregation_type=aggregation_type
    )
    self.assertEqual(tdata2.width, tdata.width)
    self.assertEqual(
        tdata2.values.shape,
        tuple([tdata2.width // resolution] * num_positional + [3]),
    )
    if aggregation_type == track_data.AggregationType.SUM:
      self.assertEqual(tdata2.values.sum(), tdata.values.sum())
    elif aggregation_type == track_data.AggregationType.MAX:
      self.assertEqual(tdata2.values.max(), tdata.values.max())


class TrackDataTrackSubsetTest(parameterized.TestCase):

  @parameterized.parameters([
      dict(num_positional=0),
      dict(num_positional=1),
      dict(num_positional=2),
  ])
  def test_filter_tracks(self, num_positional: int):
    values = (
        np.arange(2**num_positional * 3)
        .reshape([2] * num_positional + [3])
        .astype(np.int32)
    )
    metadata = pd.DataFrame({
        'name': ['first', 'first', 'second'],
        'strand': ['+', '-', '.'],
        'other': ['foo', 'foo', 'bar'],
    })
    tdata = track_data.TrackData(values, metadata, resolution=2)
    tdata2 = tdata.filter_tracks([False, True, False])
    self.assertEqual(tdata2.num_tracks, 1)
    self.assertListEqual(list(tdata2.names), ['first'])
    self.assertListEqual(list(tdata2.strands), ['-'])
    self.assertEqual(tdata2.metadata['other'].iloc[0], 'foo')
    self.assertEqual(tdata2.values.shape, tuple([2] * num_positional + [1]))

  @parameterized.parameters([
      dict(num_positional=0, strand='+'),
      dict(num_positional=1, strand='+'),
      dict(num_positional=2, strand='+'),
      dict(num_positional=0, strand='-'),
      dict(num_positional=1, strand='-'),
      dict(num_positional=2, strand='-'),
      dict(num_positional=0, strand='.'),
      dict(num_positional=1, strand='.'),
      dict(num_positional=2, strand='.'),
  ])
  def test_filter_to_strand(self, num_positional: int, strand: str):
    values = (
        np.arange(2**num_positional * 3)
        .reshape([2] * num_positional + [3])
        .astype(np.int32)
    )
    metadata = pd.DataFrame({
        'name': ['first', 'first', 'second'],
        'strand': ['+', '-', '.'],
        'other': ['foo', 'foo', 'bar'],
    })
    tdata = track_data.TrackData(values, metadata, resolution=2)
    if strand == '+':
      tdata2 = tdata.filter_to_positive_strand()
      name = 'first'
      other = 'foo'
    elif strand == '-':
      tdata2 = tdata.filter_to_negative_strand()
      name = 'first'
      other = 'foo'
    elif strand == '.':
      tdata2 = tdata.filter_to_unstranded()
      name = 'second'
      other = 'bar'
    else:
      raise ValueError(f'Invalid strand: {strand}')
    self.assertEqual(tdata2.num_tracks, 1)
    self.assertListEqual(list(tdata2.names), [name])
    self.assertListEqual(list(tdata2.strands), [strand])
    self.assertEqual(tdata2.metadata['other'].iloc[0], other)
    self.assertEqual(tdata2.values.shape, tuple([2] * num_positional + [1]))

  @parameterized.parameters([
      dict(num_positional=0),
      dict(num_positional=1),
      dict(num_positional=2),
  ])
  def test_filter_nonpositive_nonnegative_strand(self, num_positional: int):
    values = (
        np.arange(2**num_positional * 3)
        .reshape([2] * num_positional + [3])
        .astype(np.int32)
    )
    metadata = pd.DataFrame({
        'name': ['first', 'first', 'second'],
        'strand': ['+', '-', '.'],
        'other': ['foo', 'foo', 'bar'],
    })
    tdata = track_data.TrackData(values, metadata, resolution=2)
    tdata2 = tdata.filter_to_nonpositive_strand()
    self.assertListEqual(list(tdata2.strands), ['-', '.'])
    tdata2 = tdata.filter_to_nonnegative_strand()
    self.assertListEqual(list(tdata2.strands), ['+', '.'])

  @parameterized.parameters([
      dict(num_positional=0),
      dict(num_positional=1),
      dict(num_positional=2),
  ])
  def test_filter_to_stranded(self, num_positional: int):
    values = (
        np.arange(2**num_positional * 3)
        .reshape([2] * num_positional + [3])
        .astype(np.int32)
    )
    metadata = pd.DataFrame({
        'name': ['first', 'first', 'second'],
        'strand': ['+', '-', '.'],
        'other': ['foo', 'foo', 'bar'],
    })
    tdata = track_data.TrackData(values, metadata, resolution=2)
    tdata2 = tdata.filter_to_stranded()
    self.assertEqual(tdata2.num_tracks, 2)
    self.assertListEqual(list(tdata2.names), ['first', 'first'])
    self.assertListEqual(list(tdata2.strands), ['+', '-'])
    self.assertEqual(tdata2.metadata['other'].iloc[0], 'foo')
    self.assertEqual(tdata2.values.shape, tuple([2] * num_positional + [2]))

  @parameterized.parameters([
      dict(num_positional=0),
      dict(num_positional=1),
      dict(num_positional=2),
  ])
  def test_select_tracks_by_index(self, num_positional: int):
    values = (
        np.arange(2**num_positional * 3)
        .reshape([2] * num_positional + [3])
        .astype(np.int32)
    )
    metadata = pd.DataFrame({
        'name': ['first', 'first', 'second'],
        'strand': ['+', '-', '.'],
        'other': ['foo', 'foo', 'bar'],
    })
    tdata = track_data.TrackData(values, metadata, resolution=2)
    tdata2 = tdata.select_tracks_by_index([1])
    self.assertEqual(tdata2.num_tracks, 1)
    self.assertListEqual(list(tdata2.names), ['first'])
    self.assertListEqual(list(tdata2.strands), ['-'])
    self.assertEqual(tdata2.metadata['other'].iloc[0], 'foo')
    self.assertEqual(tdata2.values.shape, tuple([2] * num_positional + [1]))

  @parameterized.parameters([
      dict(num_positional=0),
      dict(num_positional=1),
      dict(num_positional=2),
  ])
  def test_select_tracks_by_name(self, num_positional: int):
    values = (
        np.arange(2**num_positional * 3)
        .reshape([2] * num_positional + [3])
        .astype(np.int32)
    )
    metadata = pd.DataFrame({
        'name': ['first', 'first', 'second'],
        'strand': ['+', '-', '.'],
        'other': ['foo', 'foo', 'bar'],
    })
    tdata = track_data.TrackData(values, metadata, resolution=2)
    tdata2 = tdata.select_tracks_by_name(['first'])
    self.assertEqual(tdata2.num_tracks, 2)
    self.assertListEqual(list(tdata2.names), ['first', 'first'])
    self.assertListEqual(list(tdata2.strands), ['+', '-'])
    self.assertEqual(tdata2.metadata['other'].iloc[0], 'foo')
    self.assertEqual(tdata2.values.shape, tuple([2] * num_positional + [2]))
    tdata2 = tdata.select_tracks_by_name(['second'])
    self.assertEqual(tdata2.num_tracks, 1)
    self.assertListEqual(list(tdata2.names), ['second'])
    self.assertListEqual(list(tdata2.strands), ['.'])
    self.assertEqual(tdata2.metadata['other'].iloc[0], 'bar')
    self.assertEqual(tdata2.values.shape, tuple([2] * num_positional + [1]))

  @parameterized.parameters([
      dict(num_positional=0),
      dict(num_positional=1),
      dict(num_positional=2),
  ])
  def test_groupby(self, num_positional: int):
    values = (
        np.arange(2**num_positional * 3)
        .reshape([2] * num_positional + [3])
        .astype(np.int32)
    )
    metadata = pd.DataFrame({
        'name': ['first', 'first', 'second'],
        'strand': ['+', '-', '.'],
        'other': ['foo', 'foo', 'bar'],
    })
    tdata = track_data.TrackData(values, metadata, resolution=2)
    tdata2_dict = tdata.groupby('other')
    self.assertListEqual(list(tdata2_dict.keys()), ['foo', 'bar'])
    self.assertListEqual(list(tdata2_dict['foo'].names), ['first', 'first'])
    self.assertListEqual(list(tdata2_dict['bar'].names), ['second'])
    self.assertEqual(tdata2_dict['foo'].num_tracks, 2)
    self.assertEqual(tdata2_dict['bar'].num_tracks, 1)
    self.assertEqual(tdata2_dict['foo'].width, tdata.width)
    self.assertEqual(tdata2_dict['bar'].width, tdata.width)


class TrackDataGetItemTest(parameterized.TestCase):

  def _get_test_data(self, num_positional: int):
    shape = [4] * num_positional + [5] if num_positional > 0 else [5]
    values = np.arange(np.prod(shape)).reshape(shape).astype(np.int32)
    metadata = pd.DataFrame({
        'name': ['a', 'b', 'c', 'c', 'd'],
        'strand': ['.', '.', '-', '+', '.'],
    })
    interval = genome.Interval('chr1', 10, 14) if num_positional > 0 else None
    return track_data.TrackData(
        values, metadata, resolution=1, interval=interval
    )

  @parameterized.parameters([
      dict(num_positional=1),
      dict(num_positional=2),
  ])
  def test_getitem_positional_slice(self, num_positional: int):
    tdata = self._get_test_data(num_positional)
    sliced_tdata = tdata[1:3]
    self.assertEqual(sliced_tdata.width, 2)
    expected_shape = tuple([2] * num_positional + [5])
    assert sliced_tdata.interval is not None
    self.assertEqual(sliced_tdata.values.shape, expected_shape)
    self.assertEqual(sliced_tdata.interval.start, 11)
    self.assertEqual(sliced_tdata.interval.end, 13)

    with self.assertRaises(IndexError):
      _ = tdata[1:3:2]

  @parameterized.parameters([
      dict(num_positional=1),
      dict(num_positional=2),
  ])
  def test_getitem_positional_interval(self, num_positional: int):
    tdata = self._get_test_data(num_positional)
    sub_interval = genome.Interval('chr1', 11, 13)
    sliced_tdata = tdata[sub_interval]
    self.assertEqual(sliced_tdata.width, 2)
    expected_shape = tuple([2] * num_positional + [5])
    assert sliced_tdata.interval is not None
    self.assertEqual(sliced_tdata.values.shape, expected_shape)
    self.assertEqual(sliced_tdata.interval, sub_interval)

  @parameterized.parameters([
      dict(num_positional=1),
      dict(num_positional=2),
  ])
  def test_getitem_positional_int(self, num_positional: int):
    tdata = self._get_test_data(num_positional)
    sliced_tdata = tdata[1]
    self.assertEqual(sliced_tdata.width, 1)
    expected_shape = tuple([1] * num_positional + [5])
    assert sliced_tdata.interval is not None
    self.assertEqual(sliced_tdata.values.shape, expected_shape)
    self.assertEqual(sliced_tdata.interval.start, 11)
    self.assertEqual(sliced_tdata.interval.end, 12)

  @parameterized.parameters([
      dict(num_positional=0),
      dict(num_positional=1),
      dict(num_positional=2),
  ])
  def test_getitem_track_slice(self, num_positional: int):
    tdata = self._get_test_data(num_positional)
    # Full slice
    sliced_tdata = tdata[:, :]
    self.assertEqual(sliced_tdata.num_tracks, 5)
    pd.testing.assert_frame_equal(sliced_tdata.metadata, tdata.metadata)
    np.testing.assert_array_equal(sliced_tdata.values, tdata.values)

    # Slice with stop
    sliced_tdata = tdata[:, :4]
    self.assertEqual(sliced_tdata.num_tracks, 4)
    self.assertListEqual(list(sliced_tdata.names), ['a', 'b', 'c', 'c'])

    # Slice with start
    sliced_tdata = tdata[:, 1:]
    self.assertEqual(sliced_tdata.num_tracks, 4)
    self.assertListEqual(list(sliced_tdata.names), ['b', 'c', 'c', 'd'])

    # Slice with start and stop
    sliced_tdata = tdata[:, 1:4]
    self.assertEqual(sliced_tdata.num_tracks, 3)
    self.assertListEqual(list(sliced_tdata.names), ['b', 'c', 'c'])
    expected_shape = list(tdata.values.shape)
    expected_shape[-1] = 3
    self.assertEqual(sliced_tdata.values.shape, tuple(expected_shape))

  def test_getitem_track_slice_with_no_positional(self):
    tdata = self._get_test_data(0)
    sliced_tdata = tdata[1:4]
    self.assertEqual(sliced_tdata.num_tracks, 3)
    self.assertListEqual(list(sliced_tdata.names), ['b', 'c', 'c'])
    expected_shape = list(tdata.values.shape)
    expected_shape[-1] = 3
    self.assertEqual(sliced_tdata.values.shape, tuple(expected_shape))

  @parameterized.parameters([
      dict(num_positional=0),
      dict(num_positional=1),
      dict(num_positional=2),
  ])
  def test_getitem_track_int(self, num_positional: int):
    tdata = self._get_test_data(num_positional)
    sliced_tdata = tdata[:, 1]
    self.assertEqual(sliced_tdata.num_tracks, 1)
    self.assertListEqual(list(sliced_tdata.names), ['b'])
    expected_shape = list(tdata.values.shape)
    expected_shape[-1] = 1
    self.assertEqual(sliced_tdata.values.shape, tuple(expected_shape))

  @parameterized.parameters([
      dict(num_positional=0),
      dict(num_positional=1),
      dict(num_positional=2),
  ])
  def test_getitem_track_indices(self, num_positional: int):
    tdata = self._get_test_data(num_positional)
    # List of ints
    sliced_tdata = tdata[:, [0, 2, 4]]
    self.assertEqual(sliced_tdata.num_tracks, 3)
    self.assertListEqual(list(sliced_tdata.names), ['a', 'c', 'd'])
    expected_shape = list(tdata.values.shape)
    expected_shape[-1] = 3
    self.assertEqual(sliced_tdata.values.shape, tuple(expected_shape))

    # Numpy array of ints
    sliced_tdata = tdata[:, np.array([0, 2, 4])]
    self.assertEqual(sliced_tdata.num_tracks, 3)
    self.assertListEqual(list(sliced_tdata.names), ['a', 'c', 'd'])
    self.assertEqual(sliced_tdata.values.shape, tuple(expected_shape))

  @parameterized.product(
      [dict(num_positional=p) for p in [0, 1, 2]],
      [
          dict(
              track_names=['a', 'd'],
              expected_names=['a', 'd'],
              expected_num_tracks=2,
          ),
          dict(
              track_names=np.array(['a', 'd']),
              expected_names=['a', 'd'],
              expected_num_tracks=2,
          ),
          dict(
              track_names=['c'],
              expected_names=['c', 'c'],
              expected_num_tracks=2,
          ),
          dict(
              track_names=np.array(['c']),
              expected_names=['c', 'c'],
              expected_num_tracks=2,
          ),
      ],
  )
  def test_getitem_track_names(
      self,
      num_positional: int,
      track_names: list[str] | np.ndarray,
      expected_names: list[str],
      expected_num_tracks: int,
  ):
    tdata = self._get_test_data(num_positional)
    sliced_tdata = tdata[:, track_names]
    self.assertEqual(sliced_tdata.num_tracks, expected_num_tracks)
    self.assertListEqual(list(sliced_tdata.names), expected_names)
    expected_shape = list(tdata.values.shape)
    expected_shape[-1] = expected_num_tracks
    self.assertEqual(sliced_tdata.values.shape, tuple(expected_shape))

  @parameterized.parameters([
      dict(num_positional=0),
      dict(num_positional=1),
      dict(num_positional=2),
  ])
  def test_getitem_track_name_str(self, num_positional: int):
    tdata = self._get_test_data(num_positional)
    # Single track
    sliced_tdata = tdata[:, 'a']
    self.assertEqual(sliced_tdata.num_tracks, 1)
    self.assertListEqual(list(sliced_tdata.names), ['a'])
    expected_shape = list(tdata.values.shape)
    expected_shape[-1] = 1
    self.assertEqual(sliced_tdata.values.shape, tuple(expected_shape))

    # Duplicated track name
    sliced_tdata = tdata[:, 'c']
    self.assertEqual(sliced_tdata.num_tracks, 2)
    self.assertListEqual(list(sliced_tdata.names), ['c', 'c'])
    expected_shape = list(tdata.values.shape)
    expected_shape[-1] = 2
    self.assertEqual(sliced_tdata.values.shape, tuple(expected_shape))

  @parameterized.parameters([
      dict(num_positional=1),
      dict(num_positional=2),
  ])
  def test_getitem_combined(self, num_positional: int):
    tdata = self._get_test_data(num_positional)
    # Positional slice and track name list
    sliced_tdata = tdata[1:3, ['a', 'c']]
    self.assertEqual(sliced_tdata.width, 2)
    self.assertEqual(sliced_tdata.num_tracks, 3)  # 'c' is duplicated
    self.assertListEqual(list(sliced_tdata.names), ['a', 'c', 'c'])
    expected_shape = tuple([2] * num_positional + [3])
    self.assertEqual(sliced_tdata.values.shape, expected_shape)
    assert sliced_tdata.interval is not None
    self.assertEqual(sliced_tdata.interval.start, 11)
    self.assertEqual(sliced_tdata.interval.end, 13)

  def test_getitem_errors(self):
    # pytype: disable=unsupported-operands
    tdata = self._get_test_data(1)
    with self.assertRaises(IndexError):
      _ = tdata[1.0]
    with self.assertRaises(IndexError):
      _ = tdata[:, 1.0]
    with self.assertRaises(IndexError):
      _ = tdata[1:4:2, :]
    with self.assertRaises(IndexError):
      _ = tdata['a']  # string is not supported for positional
    # pytype: enable=unsupported-operands

    tdata = self._get_test_data(0)
    with self.assertRaises(IndexError):
      _ = tdata[genome.Interval('chr1', 10, 14)]


class TrackDataReverseComplementTest(parameterized.TestCase):

  @parameterized.parameters([
      dict(num_positional=0),
      dict(num_positional=1),
      dict(num_positional=2),
  ])
  def test_reverse_complement(self, num_positional: int):
    values = (
        np.arange(2**num_positional * 3)
        .reshape([2] * num_positional + [3])
        .astype(np.int32)
    )
    metadata = pd.DataFrame({
        'name': ['first', 'first', 'second'],
        'strand': ['+', '-', '.'],
        'other': ['foo', 'foo', 'bar'],
    })
    tdata = track_data.TrackData(values, metadata, resolution=2)
    tdata2 = tdata.reverse_complement()
    if num_positional == 0:
      self.assertTrue(np.all(tdata2.values[[1, 0, 2]] == tdata.values))
    elif num_positional == 1:
      self.assertTrue(np.all(tdata2.values[::-1, [1, 0, 2]] == tdata.values))
    elif num_positional == 2:
      self.assertTrue(
          np.all(tdata2.values[::-1, ::-1, [1, 0, 2]] == tdata.values)
      )
    self.assertListEqual(list(tdata2.names), list(tdata.names))
    self.assertListEqual(list(tdata2.strands), ['-', '+', '.'])

  @parameterized.parameters([
      dict(num_positional=0),
      dict(num_positional=1),
      dict(num_positional=2),
  ])
  def test_reverse_complement_raises(self, num_positional: int):
    values = (
        np.arange(2**num_positional * 3)
        .reshape([2] * num_positional + [3])
        .astype(np.int32)
    )
    metadata = pd.DataFrame({
        'name': ['first', 'third', 'second'],
        'strand': ['+', '-', '.'],
        'other': ['foo', 'foo', 'bar'],
    })
    tdata = track_data.TrackData(values, metadata, resolution=2)
    with self.assertRaises(ValueError):
      # Not matching number of strands by name
      tdata.reverse_complement()


class TrackDataConcatTest(parameterized.TestCase):

  @parameterized.parameters([
      dict(num_positional=0),
      dict(num_positional=1),
      dict(num_positional=2),
      dict(num_positional=0, new_metadata_values=['tdata1', 'tdata2']),
      dict(num_positional=0, new_metadata_values=['tdata1', None]),
  ])
  def test_concat(
      self, num_positional: int, new_metadata_values: list[str] | None = None
  ):
    values = (
        np.arange(2**num_positional * 3)
        .reshape([2] * num_positional + [3])
        .astype(np.int32)
    )
    metadata1 = pd.DataFrame({
        'name': ['first', 'first', 'second'],
        'strand': ['+', '-', '.'],
        'other': ['foo', 'foo', 'bar'],
    })
    tdata = track_data.TrackData(values, metadata1, resolution=2, uns={'a': 1})
    metadata2 = pd.DataFrame({
        'name': ['third', 'third', 'fourth'],
        'strand': ['+', '-', '.'],
        'other': ['foo', 'foo', 'bar'],
    })
    tdata2 = track_data.TrackData(values, metadata2, resolution=2, uns={'a': 1})
    with self.assertRaises(ValueError):
      track_data.concat([tdata, tdata])
    tdata3 = track_data.concat(
        [tdata, tdata2],
        ('new_column', new_metadata_values) if new_metadata_values else None,
    )
    self.assertEqual(tdata3.num_tracks, 6)
    self.assertEqual(tdata3.width, tdata.width)
    self.assertIsNone(tdata3.uns)
    self.assertEqual(tdata3.values.shape, tuple([2] * num_positional + [6]))
    self.assertLen(tdata3.metadata, 6)
    if new_metadata_values:
      # Expect error if the column already exists.
      with self.assertRaises(ValueError):
        track_data.concat([tdata, tdata2], ('name', new_metadata_values))
      self.assertSequenceEqual(
          list(tdata3.metadata['new_column'].unique()), new_metadata_values
      )
    else:
      self.assertNotIn('new_column', tdata3.metadata.columns)


class TrackDataInterleaveTest(absltest.TestCase):

  def test_interleave(self):
    values_1 = np.array(
        [
            [0.0, 0.2, 0.3, 0.4, 0.5],
            [0.1, 0.3, 0.5, 0.6, 0.7],
            [0.2, 0.4, 0.7, 0.8, 0.9],
        ],
        dtype=np.float32,
    )
    metadata_1 = pd.DataFrame({
        'name': ['track_0', 'track_1', 'track_2', 'track_3', 'track_4'],
        'strand': '.',
        'gtex_tissue': ['BLADDER', 'LIVER', 'LIVER', 'STOMACH', 'STOMACH'],
        'padding': [False, False, False, False, True],
    })
    values_2 = -1 * values_1
    metadata_2 = metadata_1.copy()
    data_1 = track_data.TrackData(values=values_1, metadata=metadata_1)
    data_2 = track_data.TrackData(values=values_2, metadata=metadata_2)
    track_datas = [data_1, data_2]
    name_prefixes = ['REF_', 'ALT_']
    interleaved_track_data = track_data.interleave(track_datas, name_prefixes)
    expected_values = np.array(
        [
            [0.0, -0.0, 0.2, -0.2, 0.3, -0.3, 0.4, -0.4, 0.5, -0.5],
            [0.1, -0.1, 0.3, -0.3, 0.5, -0.5, 0.6, -0.6, 0.7, -0.7],
            [0.2, -0.2, 0.4, -0.4, 0.7, -0.7, 0.8, -0.8, 0.9, -0.9],
        ],
        dtype=np.float32,
    )
    np.testing.assert_array_equal(
        interleaved_track_data.values, expected_values
    )
    expected_metadata = pd.DataFrame({
        'name': [
            'REF_track_0',
            'ALT_track_0',
            'REF_track_1',
            'ALT_track_1',
            'REF_track_2',
            'ALT_track_2',
            'REF_track_3',
            'ALT_track_3',
            'REF_track_4',
            'ALT_track_4',
        ],
        'strand': ['.', '.', '.', '.', '.', '.', '.', '.', '.', '.'],
        'gtex_tissue': [
            'BLADDER',
            'BLADDER',
            'LIVER',
            'LIVER',
            'LIVER',
            'LIVER',
            'STOMACH',
            'STOMACH',
            'STOMACH',
            'STOMACH',
        ],
        'padding': [
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            True,
            True,
        ],
    })
    pd.testing.assert_frame_equal(
        interleaved_track_data.metadata, expected_metadata
    )


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest
from absl.testing import parameterized
from alphagenome.data import gene_annotation
from alphagenome.data import genome
import numpy as np
import pandas as pd


class ExtractTSSTest(absltest.TestCase):

  def test_extract_tss(self):
    gtf = pd.DataFrame({
        'Start': [101, 102, 103, 0],
        'End': [200, 200, 200, 107],
        'Strand': ['+', '+', '+', '-'],
        'transcript_id': ['T1', 'T2', 'T2', 'T4'],
    })
    gtf['Feature'] = 'transcript'
    tss = gene_annotation.extract_tss(gtf)
    self.assertSequenceEqual(list(tss.Start), [101, 102, 103, 107])
    self.assertSequenceEqual(list(tss.End), [101, 102, 103, 107])
    self.assertSequenceEqual(list(tss.Strand), list(gtf.Strand))
    self.assertSequenceEqual(list(tss.transcript_id), list(gtf.transcript_id))


class UpgradeAnnotationIdsTest(absltest.TestCase):

  def test_upgrade_annotation_ids(self):
    old_ids = ['T1.1', 'T2.2', 'T3.3', 'T3.3_PAR_Y']
    new_ids = ['T0.3', 'T1.2', 'T2.2', 'T3.1', 'T3.4_PAR_Y', 'T4.2']
    upgrade_ids = gene_annotation.upgrade_annotation_ids(
        pd.Series(old_ids), pd.Series(new_ids)
    )
    self.assertSequenceEqual(
        list(upgrade_ids), ['T1.2', 'T2.2', 'T3.1', 'T3.4_PAR_Y']
    )

  def test_upgrade_annotation_ids_missing(self):
    old_ids = ['T1.1', 'T2.2', 'T3.3_PAR_Y']
    new_ids = ['T1.2', 'T2.2']
    upgrade_ids = gene_annotation.upgrade_annotation_ids(
        pd.Series(old_ids), pd.Series(new_ids)
    )
    self.assertSequenceEqual(list(upgrade_ids), ['T1.2', 'T2.2', np.nan])

  def test_upgrade_annotation_ids_missing_patch(self):
    old_ids = ['T1.1', 'T2', 'T3.3_PAR_Y']
    new_ids = ['T1.2', 'T2.2']
    with self.assertRaises(ValueError):
      gene_annotation.upgrade_annotation_ids(
          pd.Series(old_ids), pd.Series(new_ids)
      )

  def test_upgrade_annotation_ids_not_unique(self):
    old_ids = ['T1.1', 'T2.1', 'T2.2', 'T3.3_PAR_Y']
    new_ids = ['T1.2', 'T2.2']
    with self.assertRaises(AssertionError):
      gene_annotation.upgrade_annotation_ids(
          pd.Series(old_ids), pd.Series(new_ids)
      )

  def test_upgrade_annotation_ids_patchless(self):
    old_ids = ['T1', 'T2', 'T3', 'T3PAR_Y']
    new_ids = ['T0.3', 'T1.2', 'T2.2', 'T3.1', 'T3.4_PAR_Y', 'T4.2']
    upgrade_ids = gene_annotation.upgrade_annotation_ids(
        pd.Series(old_ids), pd.Series(new_ids), True
    )
    self.assertSequenceEqual(
        list(upgrade_ids), ['T1.2', 'T2.2', 'T3.1', 'T3.4_PAR_Y']
    )


class GetGeneIntervalTest(parameterized.TestCase):

  @parameterized.parameters([
      dict(gene_symbol='TP53'),
      dict(gene_symbol='tp53'),  # Should handle gene symbol case differences.
      dict(gene_symbol='EGFR'),
      dict(gene_symbol='TNF'),
      dict(gene_id='ENSG01.1'),
      dict(gene_id='ENSG01'),
  ])
  def test_get_gene_interval(self, gene_symbol=None, gene_id=None):
    gtf = pd.DataFrame({
        'gene_id': ['ENSG01.1', 'ENSG02.2', 'ENSG03.1'],
        'gene_name': ['TP53', 'EGFR', 'TNF'],
        'Chromosome': ['chr1', 'chr2', 'chr3'],
        'Start': [101, 150, 101],
        'End': [200, 200, 200],
        'Strand': ['+', '+', '-'],
        'Feature': 'gene',
    })
    interval = gene_annotation.get_gene_interval(
        gtf=gtf, gene_symbol=gene_symbol, gene_id=gene_id
    )
    self.assertIsNotNone(interval)

  @parameterized.parameters([
      dict(
          gene_symbol='FOO',
          gene_id=None,
          expected_error=r'No interval found for gene\(s\): FOO.',
      ),
      dict(
          gene_id='BAR',
          gene_symbol=None,
          expected_error=r'No interval found for gene\(s\): BAR.',
      ),
      dict(
          gene_id=None,
          gene_symbol=None,
          expected_error='Exactly one of gene_symbol or gene_id must be set.',
      ),
      dict(
          gene_id='ENSG01.1',
          gene_symbol=None,
          expected_error=r'Multiple intervals found for gene\(s\): ENSG01.1.',
      ),
  ])
  def test_invalid_gene_raises_error(
      self, gene_symbol, gene_id, expected_error
  ):
    gtf = pd.DataFrame({
        'gene_id': ['ENSG01.1', 'ENSG01.1', 'ENSG03.1'],
        'gene_name': ['TP53', 'EGFR', 'TNF'],
        'Chromosome': ['chr1', 'chr2', 'chr3'],
        'Start': [101, 150, 101],
        'End': [200, 200, 200],
        'Strand': ['+', '+', '-'],
        'Feature': 'gene',
    })
    with self.assertRaisesRegex(ValueError, expected_error):
      gene_annotation.get_gene_interval(
          gtf=gtf, gene_symbol=gene_symbol, gene_id=gene_id
      )


class GetGeneIntervalsTest(parameterized.TestCase):

  @parameterized.parameters([
      dict(
          gene_symbols=['TP53', 'EGFR', 'TNF'],
          expected_intervals=[
              genome.Interval('chr1', 101, 200, '+', 'TP53'),
              genome.Interval('chr2', 150, 200, '+', 'EGFR'),
              genome.Interval('chr3', 101, 200, '-', 'TNF'),
          ],
      ),
      dict(
          gene_symbols=['tp53', 'EGFR'],  # Should handle case differences.
          expected_intervals=[
              genome.Interval('chr1', 101, 200, '+', 'tp53'),
              genome.Interval('chr2', 150, 200, '+', 'EGFR'),
          ],
      ),
      dict(
          gene_ids=['ENSG01', 'ENSG03.1'],  # Should handle missing patch.
          expected_intervals=[
              genome.Interval('chr1', 101, 200, '+', 'ENSG01'),
              genome.Interval('chr3', 101, 200, '-', 'ENSG03.1'),
          ],
      ),
      dict(
          gene_symbols=['TNF', 'TP53'],  # Test order is preserved.
          expected_intervals=[
              genome.Interval('chr3', 101, 200, '-', 'TNF'),
              genome.Interval('chr1', 101, 200, '+', 'TP53'),
          ],
      ),
      dict(
          gene_symbols=['TP53', 'EGFR', 'TP53'],  # Test duplicates in input.
          expected_intervals=[
              genome.Interval('chr1', 101, 200, '+', 'TP53'),
              genome.Interval('chr2', 150, 200, '+', 'EGFR'),
              genome.Interval('chr1', 101, 200, '+', 'TP53'),
          ],
      ),
  ])
  def test_get_gene_intervals(
      self, expected_intervals, gene_symbols=None, gene_ids=None
  ):
    gtf = pd.DataFrame({
        'gene_id': ['ENSG01.1', 'ENSG02.2', 'ENSG03.1'],
        'gene_name': ['TP53', 'EGFR', 'TNF'],
        'Chromosome': ['chr1', 'chr2', 'chr3'],
        'Start': [101, 150, 101],
        'End': [200, 200, 200],
        'Strand': ['+', '+', '-'],
        'Feature': 'gene',
    })
    intervals = gene_annotation.get_gene_intervals(
        gtf=gtf, gene_symbols=gene_symbols, gene_ids=gene_ids
    )
    self.assertSequenceEqual(intervals, expected_intervals)

  @parameterized.parameters([
      dict(
          gene_symbols=['FOO', 'BAR'],
          gene_ids=None,
          expected_error=r'No interval found for gene\(s\): BAR, FOO.',
      ),
      dict(
          gene_ids=['ENSG01.1', 'ENSG03'],
          gene_symbols=None,
          expected_error=(
              r'Multiple intervals found for gene\(s\): ENSG01.1, ENSG03.'
          ),
      ),
  ])
  def test_invalid_genes_raise_error(
      self, gene_symbols, gene_ids, expected_error
  ):
    gtf = pd.DataFrame({
        'gene_id': ['ENSG01.1', 'ENSG01.1', 'ENSG03.1', 'ENSG03.1'],
        'gene_name': ['TP53', 'EGFR', 'TNF', 'CD8A'],
        'Chromosome': ['chr1', 'chr2', 'chr3', 'chr4'],
        'Start': [101, 150, 101, 101],
        'End': [200, 200, 200, 200],
        'Strand': ['+', '+', '-', '+'],
        'Feature': 'gene',
    })
    with self.assertRaisesRegex(ValueError, expected_error):
      gene_annotation.get_gene_intervals(
          gtf=gtf, gene_symbols=gene_symbols, gene_ids=gene_ids
      )


class FilterGTFTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    # Gene A has 3 transcripts, A1, A2, A3. A1 and A3 are the same length but
    # the first should be returned.
    # Gene B has 1 transcript, B1. It is the max length transcript due to being
    # the only one.
    # Gene C has 2 transcripts, C1, C2. C1 is the max length transcript.
    self.gtf = pd.DataFrame({
        'gene_id': ['A', 'A', 'A', 'B', 'C', 'C'],
        'transcript_id': ['A1', 'A2', 'A3', 'B1', 'C1', 'C2'],
        'Start': [101, 150, 101, 300, 500, 515],
        'End': [200, 200, 200, 420, 560, 560],
        'Strand': ['+', '+', '+', '-', '-', '-'],
        'Feature': 'transcript',
        'transcript_support_level': ['1', '2', '3', '1', '1', '2'],
        'tag': [
            'basic,MANE_Select',
            pd.NA,
            'basic,Ensembl_canonical',
            pd.NA,
            pd.NA,
            pd.NA,
        ],
    })

  def test_filter_to_longest_transcript(self):
    longest_transcript_gtf = gene_annotation.filter_to_longest_transcript(
        self.gtf
    )
    self.assertSequenceEqual(
        list(longest_transcript_gtf.transcript_id), ['A1', 'B1', 'C1']
    )

  def test_filter_to_mane_select_transcript(self):
    mane_select_gtf = gene_annotation.filter_to_mane_select_transcript(self.gtf)
    self.assertSequenceEqual(list(mane_select_gtf.transcript_id), ['A1'])

  def test_filter_to_mane_select_transcript_raises_error(self):
    with self.assertRaises(ValueError):
      gene_annotation.filter_to_mane_select_transcript(
          self.gtf[self.gtf['tag'].isna()]
      )

  @parameterized.parameters([
      dict(
          support_level=['1'],
          expected=['A1', 'B1', 'C1'],
      ),
      dict(
          support_level='1',
          expected=['A1', 'B1', 'C1'],
      ),
      dict(
          support_level=['1', '2'],
          expected=['A1', 'A2', 'B1', 'C1', 'C2'],
      ),
  ])
  def test_filter_transcript_support_level(self, support_level, expected):
    support_level1_gtf = gene_annotation.filter_transcript_support_level(
        self.gtf, support_level
    )
    self.assertSequenceEqual(list(support_level1_gtf.transcript_id), expected)

  def test_filter_transcript_support_level_invalid(self):
    # Test expected error if any invalid support level is requested.
    with self.assertRaises(ValueError):
      gene_annotation.filter_transcript_support_level(self.gtf, ['1', '6'])


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for working with gene annotations (e.g., GTFs)."""

from collections.abc import Sequence
import enum

from alphagenome.data import genome
import numpy as np
import pandas as pd


@enum.unique
class TranscriptType(enum.Enum):
  """Valid Transcript types available in the GENCODE GTF."""

  IG_C_GENE = 'IG_C_gene'
  IG_C_PSEUDOGENE = 'IG_C_pseudogene'
  IG_D_GENE = 'IG_D_gene'
  IG_J_GENE = 'IG_J_gene'
  IG_J_PSEUDOGENE = 'IG_J_pseudogene'
  IG_V_GENE = 'IG_V_gene'
  IG_V_PSEUDOGENE = 'IG_V_pseudogene'
  IG_PSEUDOGENE = 'IG_pseudogene'
  MT_RRNA = 'Mt_rRNA'
  MT_TRNA = 'Mt_tRNA'
  TEC = 'TEC'
  TR_C_GENE = 'TR_C_gene'
  TR_D_GENE = 'TR_D_gene'
  TR_J_GENE = 'TR_J_gene'
  TR_J_PSEUDOGENE = 'TR_J_pseudogene'
  TR_V_GENE = 'TR_V_gene'
  TR_V_PSEUDOGENE = 'TR_V_pseudogene'
  ARTIFACT = 'artifact'
  LNCRNA = 'lncRNA'
  MIRNA = 'miRNA'
  MISC_RNA = 'misc_RNA'
  NON_STOP_DECAY = 'non_stop_decay'
  NONSENSE_MEDIATED_DECAY = 'nonsense_mediated_decay'
  PROCESSED_PSEUDOGENE = 'processed_pseudogene'
  PROCESSED_TRANSCRIPT = 'processed_transcript'
  PROTEIN_CODING = 'protein_coding'
  PROTEIN_CODING_CDS_NOT_DEFINED = 'protein_coding_CDS_not_defined'
  PROTEIN_CODING_LOF = 'protein_coding_LoF'
  RRNA = 'rRNA'
  RRNA_PSEUDOGENE = 'rRNA_pseudogene'
  RETAINED_INTRON = 'retained_intron'
  RIBOZYME = 'ribozyme'
  SRNA = 'sRNA'
  SCRNA = 'scRNA'
  SCARNA = 'scaRNA'
  SNRNA = 'snRNA'
  SNORNA = 'snoRNA'
  TRANSCRIBED_PROCESSED_PSEUDOGENE = 'transcribed_processed_pseudogene'
  TRANSCRIBED_UNITARY_PSEUDOGENE = 'transcribed_unitary_pseudogene'
  TRANSCRIBED_UNPROCESSED_PSEUDOGENE = 'transcribed_unprocessed_pseudogene'
  TRANSLATED_PROCESSED_PSEUDOGENE = 'translated_processed_pseudogene'
  UNITARY_PSEUDOGENE = 'unitary_pseudogene'
  UNPROCESSED_PSEUDOGENE = 'unprocessed_pseudogene'
  VAULT_RNA = 'vault_RNA'


def extract_tss(gtf: pd.DataFrame, feature: str = 'transcript') -> pd.DataFrame:
  """Extract transcription start sites (TSS) from a DataFrame.

  Args:
    gtf: pd.DataFrame containing gene annotation.
    feature: Feature in the GTF file to use (either transcript or gene).

  Returns:
    pd.DataFrame containing transcription start sites (width=0, 0-based).
  """
  tss = gtf[(gtf.Feature == feature)].copy()

  # Remove the extra base to make it width=0.
  # .....[)TRANSCRIPT (strand = +)
  # TPIRCSNART[)..... (strand = -)
  new_start = np.where(tss.Strand == '-', tss.End, tss.Start)
  tss.Start = new_start
  tss.End = new_start

  return tss


def filter_transcript_type(
    gtf: pd.DataFrame,
    transcript_types: tuple[TranscriptType, ...] | None = None,
) -> pd.DataFrame:
  """Filter GTF entries by transcript types.

  This function takes a GTF DataFrame and a list of transcript types and returns
  a new DataFrame containing only the transcripts with the specified types.

  The GTF DataFrame must contain a column named 'transcript_type' or
  'transcript_biotype'. The function will raise a ValueError if neither of these
  columns is present.

  Args:
    gtf: pd.DataFrame or pyranges.PyRanges.
    transcript_types: List of valid transcript types to use for filtering.

  Returns:
    pd.DataFrame of GENCODE GTF entries subset to rows with the requested
    transcript types.
  """
  if transcript_types is not None:
    transcript_types_str = [x.value for x in transcript_types]
    if 'transcript_type' in gtf.columns:
      gtf = gtf[gtf.transcript_type.isin(transcript_types_str)]
    elif 'transcript_biotype' in gtf.columns:
      gtf = gtf[gtf.transcript_biotype.isin(transcript_types_str)]
    else:
      raise ValueError('transcript_type or transcript_biotype not in gtf.')
  return gtf


def filter_protein_coding(
    gtf: pd.DataFrame, include_gene_entries: bool = False
) -> pd.DataFrame:
  """Filter GTF entries to only protein-coding genes.

  Args:
    gtf: pd.DataFrame of GENCODE GTF entries. This data frame must contain a
      column named 'transcript_type' or 'transcript_biotype'.
    include_gene_entries: Whether to include gene entries in addition to
      transcript entries.

  Returns:
    pd.DataFrame of GENCODE GTF entries subset to rows with protein-coding
    genes.
  """
  if include_gene_entries:
    if 'gene_type' in gtf.columns:
      gtf = gtf[gtf.gene_type == TranscriptType.PROTEIN_CODING.value]
    else:
      raise ValueError('gene_type not in gtf.')
  else:
    gtf = filter_transcript_type(gtf, (TranscriptType.PROTEIN_CODING,))
  return gtf


def filter_to_longest_transcript(
    gtf: pd.DataFrame,
) -> pd.DataFrame:
  """Filter GTF entries to only the longest transcript per gene.

  Args:
    gtf: pd.DataFrame of GENCODE GTF entries. Must contain columns 'Feature',
      'End', 'Start', 'gene_id', and 'transcript_id'.

  Returns:
    pd.DataFrame of GENCODE GTF entries subset to rows with the longest
    transcript per gene.
  """
  lengths = gtf[gtf['Feature'] == 'transcript'].reset_index(drop=True)
  lengths['transcript_length'] = lengths['End'] - lengths['Start'] + 1

  # Identify longest transcripts per gene_id.
  longest_transcripts = lengths.loc[
      lengths.groupby('gene_id')['transcript_length'].idxmax()
  ]

  return gtf[gtf['transcript_id'].isin(longest_transcripts['transcript_id'])]


def filter_to_mane_select_transcript(gtf: pd.DataFrame) -> pd.DataFrame:
  """Filter GTF entries to only the MANE select transcript.

  Note that the MANE_Select tag only exists for the human GTF file.

  Args:
    gtf: pd.DataFrame of GENCODE GTF entries. Must contain columns 'tag'.

  Returns:
    pd.DataFrame of GENCODE GTF entries subset to rows representing MANE
    select transcripts, which are transcripts that are well-supported,
    conserved, and expressed.
  """
  filtered_gtf = gtf[gtf['tag'].fillna('').str.contains('MANE_Select')]
  if filtered_gtf.empty:
    raise ValueError(
        'No MANE_Select transcripts found in the GTF, possibly due to non-human'
        ' GTF.'
    )
  return filtered_gtf


def filter_transcript_support_level(
    gtf: pd.DataFrame,
    transcript_support_levels: str | Sequence[str],
) -> pd.DataFrame:
  """Filter GTF to only transcripts with specific GENCODE support levels.

  As documented in the [Ensembl
  glossary](https://www.ensembl.org/Help/Glossary),
  the transcript support level (TSL) indicates the degree of evidence that was
  used to construct the transcript.

  As taken from the glossary, the levels are:

  | Transcript support level | Description |  |
  |---|---|---|
  | 1 | A transcript where all splice junctions are supported by at least one
  non-suspect mRNA. |
  | 2 | A transcript where the best supporting mRNA is flagged as suspect or the
  support is from multiple ESTs |
  | 3 | A transcript where the only support is from a single EST |
  | 4 | A transcript where the best supporting EST is flagged as suspect |
  | 5 | A transcript where no single transcript supports the model structure. |
  | NA | A transcript that was not analysed for TSL. |

  Args:
    gtf: pd.DataFrame of GENCODE GTF entries. Must contain column
      'transcript_support_level'.
    transcript_support_levels: List of valid transcript support levels to use
      for filtering. This must be a subset of ['1', '2', '3', '4', '5']. Can
      also be single string.

  Returns:
    pd.DataFrame exactly as provided, but subset to rows with the specified
    support level(s).

  Transcripts are scored by GENCODE according to how well mRNA and EST
  alignments
  match over its full length. Valid levels are:
  '1': All splice junctions of the transcript are supported by at least one
  non-suspect mRNA.
  '2': The best supporting mRNA is flagged as suspect or the support is from
  multiple ESTs.
  '3': The only support is from a single EST.
  '4': The best supporting EST is flagged as suspect.
  '5': No single transcript supports the model structure.
  'NA': The transcript was not analyzed (not supported by this filter function).

  See GENCODE GTF format documentation for further details:
  https://www.gencodegenes.org/pages/data_format.html
  """
  if isinstance(transcript_support_levels, str):
    transcript_support_levels = list(transcript_support_levels)

  supported_tsls = {'1', '2', '3', '4', '5'}
  if not set(transcript_support_levels).issubset(supported_tsls):
    raise ValueError(
        f'transcript_support_level must be one of {supported_tsls}, but was'
        f' {transcript_support_levels}'
    )
  return gtf[gtf.transcript_support_level.isin(transcript_support_levels)]


def upgrade_annotation_ids(
    old_ids: pd.Series, new_ids: pd.Series, patchless: bool = False
) -> pd.Series:
  """Upgrade or add transcript id patch version to Ensembl IDs.

  This function works by

  1. Dropping the patch version from `old_ids` and `new_ids`
  2. Merging the two on the patch-less ids.
  3. Returning the result of the merge as a pd.Series, with the 'old_ids' as
  the index and the 'new_ids' as the values.

  Ensembl patch versions have two formats: ENST####.<patch>_PAR_Y or
  ENST####.<patch>. Both are handled.

  The function will raise a ValueError if either `old_ids` or `new_ids` result
  in duplicates after dropping the patch version.

  Examples:
  * If the old ids are ENST00010.1 and the new ids are ENST00010.3,
  then the mapping will be ENST00010.1 -> ENST00010.3.
  * If the old ids are ENST00010 and the new ids are ENST00010.3, then the
  mapping will be ENST00010 -> ENST00010.3.

  Args:
    old_ids: A pd.Series of Ensembl transcript or gene ids with older or missing
      version/patch numbers. The index of the series is ignored.
    new_ids: A pd.Series of transcript or gene ids with newer version/patch
      numbers. The index of the series is ignored.
    patchless: whether old_ids are missing patch.

  Returns:
    A pd.Series, with the 'old_ids' as the index and the 'new_ids' as the
    values.
  """
  new_ids = new_ids.drop_duplicates()

  def drop_version(x):
    """Drop the patch version from an Ensembl ID."""
    if not x.str.contains('.', regex=False).all():
      raise ValueError('All ids need to contain the patch version.')

    id_split = x.str.partition('.')

    # Retain anything after _ such as PAR_Y for ENST####.<patch>_PAR_Y.
    return id_split[0] + id_split[2].str.partition('_')[2]

  old_ids_nopatch = old_ids if patchless else drop_version(old_ids)
  new_ids_nopatch = drop_version(new_ids)
  assert (
      not old_ids_nopatch.duplicated().any()
  ), 'old_ids not unique without version'
  assert (
      not new_ids_nopatch.duplicated().any()
  ), 'new_ids not unique without version'
  df = pd.merge(
      pd.DataFrame({'old': old_ids.values, 'no_version': old_ids_nopatch}),
      pd.DataFrame({
          'new': new_ids.values,
          'no_version': new_ids_nopatch,
      }),
      on='no_version',
      how='left',
  )
  return pd.Series(
      df.set_index('old').loc[old_ids].new.values, index=old_ids.index
  )


def get_gene_intervals(
    gtf: pd.DataFrame,
    gene_symbols: Sequence[str] | None = None,
    gene_ids: Sequence[str] | None = None,
) -> list[genome.Interval]:
  """Returns a list of stranded `genome.Interval`s for the given identifiers.

  Args:
    gtf: pd.DataFrame of GENCODE GTF entries. Must contain columns 'Feature',
      'gene_name', 'gene_id', 'Chromosome', 'Start', 'End', and 'Strand'.
    gene_symbols: A sequence of gene names or gene symbols (e.g., ['EGFR',
      'TNF', 'TP53']). Matching is case-insensitive.
    gene_ids: A sequence of Ensembl gene IDs, which can be patched (e.g.
      ['ENSG00000141510.17']) or unpatched (e.g., ['ENSG00000141510']). Matching
      is done on unpatched IDs.

  Returns:
    A list of `genome.Interval`s for the given identifiers. The
    returned list of intervals is in the same order as the input gene
    identifiers.

  Raises:
    ValueError: If neither or both gene_symbols and gene_ids are set, or if no
      interval or multiple intervals are found for any of the given gene
      identifiers.
  """
  if (gene_symbols is None) == (gene_ids is None):
    raise ValueError('Exactly one of gene_symbols or gene_ids must be set.')

  gtf_genes = gtf[gtf['Feature'] == 'gene'].copy()

  if gene_symbols is not None:
    id_col = 'gene_name'
    input_ids = gene_symbols
    process_fn = lambda s: s.str.upper()
  else:
    id_col = 'gene_id'
    input_ids = gene_ids
    process_fn = lambda s: s.str.split('.', n=1).str[0]

  processed_input_ids = process_fn(pd.Series(input_ids, dtype=str))
  gtf_genes['processed_id'] = process_fn(gtf_genes[id_col])

  # Filter the GTF to only the genes that are in the input IDs.
  gtf_subset = gtf_genes[
      gtf_genes['processed_id'].isin(processed_input_ids.unique())
  ]

  dup_mask = gtf_subset['processed_id'].duplicated(keep=False)
  if dup_mask.any():
    offending_ids = gtf_subset.loc[dup_mask, id_col].unique()
    raise ValueError(
        'Multiple intervals found for gene(s):'
        f' {", ".join(sorted(offending_ids))}.'
    )

  # Create a lookup map from processed_id to GTF data.
  # Use reindex to order genes by input and insert NaNs for missing genes.
  gtf_map = gtf_subset.set_index('processed_id')
  result_df = gtf_map.reindex(processed_input_ids)

  missing_mask = result_df['Chromosome'].isnull()
  if missing_mask.any():
    missing_ids = pd.Series(input_ids)[missing_mask.values].unique()
    raise ValueError(
        f'No interval found for gene(s): {", ".join(sorted(missing_ids))}.'
    )

  # Add original identifiers to the result for the 'name' field of the interval.
  result_df[id_col] = list(input_ids)

  return [
      genome.Interval(
          chromosome=row.Chromosome,
          start=row.Start,
          end=row.End,
          strand=row.Strand,
          name=getattr(row, id_col),
      )
      for row in result_df.itertuples()
  ]


def get_gene_interval(
    gtf: pd.DataFrame,
    gene_symbol: str | None = None,
    gene_id: str | None = None,
) -> genome.Interval:
  """Returns a stranded `genome.Interval` given a gene identifier.

  Either gene_symbol or gene_id must be set, but not both.

  Args:
    gtf: pd.DataFrame of GENCODE GTF entries. Must contain columns 'Feature',
      'gene_name', 'gene_id', 'Chromosome', 'Start', 'End', and 'Strand'.
    gene_symbol: A gene name or gene symbol (e.g., 'EGFR', 'TNF', 'TP53')
    gene_id: An Ensembl gene ID, which can be patched (e.g.
      'ENSG00000141510.17') or unpatched (e.g., 'ENSG00000141510').

  Returns:
    A `genome.Interval` for the given gene identifier.

  Raises:
    ValueError: If neither or both gene_symbol and gene_id are set, or if no
      interval or multiple intervals are found for the given gene identifier.
  """
  if sum(x is not None for x in [gene_symbol, gene_id]) != 1:
    raise ValueError('Exactly one of gene_symbol or gene_id must be set.')

  return get_gene_intervals(
      gtf,
      [gene_symbol] if gene_symbol else None,
      [gene_id] if gene_id else None,
  )[0]



# Copyright 2025 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Track data utils for converting to/from protos."""

from collections.abc import Iterable, Sequence

from alphagenome import tensor_utils
from alphagenome.data import genome
from alphagenome.data import ontology
from alphagenome.data import track_data
from alphagenome.protos import dna_model_pb2
from alphagenome.protos import tensor_pb2
import pandas as pd


def to_protos(
    data: track_data.TrackData,
    *,
    bytes_per_chunk: int = 0,
    compression_type: tensor_pb2.CompressionType = (
        tensor_pb2.CompressionType.COMPRESSION_TYPE_NONE
    ),
) -> tuple[dna_model_pb2.TrackData, Sequence[tensor_pb2.TensorChunk]]:
  """Serializes `TrackData` to protobuf messages.

  Args:
    data: The `TrackData` object to serialize.
    bytes_per_chunk: The maximum number of bytes per tensor chunk.
    compression_type: The compression type to use for the tensor chunks.

  Returns:
    A tuple containing the `TrackData` protobuf message and a sequence of
    `TensorChunk` protobuf messages.
  """
  tensor, chunks = tensor_utils.pack_tensor(
      data.values,
      bytes_per_chunk=bytes_per_chunk,
      compression_type=compression_type,
  )
  return (
      dna_model_pb2.TrackData(
          values=tensor,
          metadata=metadata_to_proto(data.metadata).metadata,
          resolution=data.resolution,
          interval=data.interval.to_proto() if data.interval else None,
      ),
      chunks,
  )


def from_protos(
    proto: dna_model_pb2.TrackData,
    chunks: Iterable[tensor_pb2.TensorChunk] = (),
    *,
    interval: genome.Interval | None = None,
) -> track_data.TrackData:
  """Creates a `TrackData` object from protobuf messages.

  Args:
    proto: A `TrackData` protobuf message.
    chunks: A sequence of `TensorChunk` protobuf messages.
    interval: Optional `Interval` object representing the genomic region
      containing the tracks. Only used if the proto does not have an interval.

  Returns:
    A `TrackData` object.
  """
  metadata = metadata_from_proto(
      dna_model_pb2.TracksMetadata(metadata=proto.metadata)
  )

  values = tensor_utils.unpack_proto(proto.values, chunks)
  values = tensor_utils.upcast_floating(values)
  resolution = proto.resolution if proto.HasField('resolution') else 1
  if proto.HasField('interval'):
    interval = genome.Interval.from_proto(proto.interval)

  return track_data.TrackData(
      values, metadata, resolution=resolution, interval=interval
  )


def metadata_to_proto(
    metadata: track_data.TrackMetadata,
) -> dna_model_pb2.TracksMetadata:
  """Converts track metadata to a `TracksMetadata` protobuf message.

  Args:
    metadata: A pandas DataFrame containing track metadata, with the following
      required columns: name, strand.

  Returns:
    A `TracksMetadata` protobuf message.
  """
  names = metadata['name']
  default_values = [None] * len(names)

  columns = zip(
      metadata['name'],
      metadata['strand'],
      metadata.get('ontology_curie', default_values),
      metadata.get('biosample_type', default_values),
      metadata.get('biosample_name', default_values),
      metadata.get('biosample_life_stage', default_values),
      metadata.get('transcription_factor', default_values),
      metadata.get('histone_mark', default_values),
      metadata.get('gtex_tissue', default_values),
      metadata.get('Assay title', default_values),
      metadata.get('data_source', default_values),
      metadata.get('genetically_modified', default_values),
      metadata.get('endedness', default_values),
      metadata.get('nonzero_mean', default_values),
      strict=True,
  )

  metadata_protos = []
  for (
      name,
      strand,
      ontology_curie,
      biosample_type,
      biosample_name,
      biosample_life_stage,
      transcription_factor,
      histone_mark,
      gtex_tissue,
      assay,
      data_source,
      genetically_modified,
      endedness,
      nonzero_mean,
  ) in columns:
    if biosample_type:
      biosample = dna_model_pb2.Biosample(
          name=biosample_name,
          type=dna_model_pb2.BiosampleType.Value(
              f'BIOSAMPLE_TYPE_{biosample_type.upper()}'
          ),
          stage=biosample_life_stage,
      )
    else:
      biosample = None
    if endedness is not None:
      match endedness:
        case 'paired':
          endedness = dna_model_pb2.Endedness.ENDEDNESS_PAIRED
        case 'single':
          endedness = dna_model_pb2.Endedness.ENDEDNESS_SINGLE
        case _:
          raise ValueError(f'Unknown endedness: {endedness}')

    metadata_protos.append(
        dna_model_pb2.TrackMetadata(
            name=name,
            strand=genome.Strand.from_str(strand).to_proto()
            if strand
            else None,
            ontology_term=ontology.from_curie(ontology_curie).to_proto()
            if ontology_curie
            else None,
            biosample=biosample,
            transcription_factor_code=transcription_factor
            if isinstance(transcription_factor, str)
            else None,
            histone_mark_code=histone_mark
            if isinstance(histone_mark, str)
            else None,
            gtex_tissue=gtex_tissue,
            assay=assay,
            data_source=data_source,
            genetically_modified=genetically_modified,
            endedness=endedness,
            nonzero_mean=nonzero_mean,
        )
    )

  return dna_model_pb2.TracksMetadata(metadata=metadata_protos)


def metadata_from_proto(
    proto: dna_model_pb2.TracksMetadata,
) -> track_data.TrackMetadata:
  """Creates track metadata from a `TracksMetadata` protobuf message.

  Args:
    proto: A `TracksMetadata` protobuf message.

  Returns:
    A pandas DataFrame containing track metadata.
  """
  metadata = []
  for track_proto in proto.metadata:
    track_metadata = {
        'name': track_proto.name,
        'strand': str(genome.Strand.from_proto(track_proto.strand)),
    }

    if track_proto.HasField('assay'):
      track_metadata['Assay title'] = track_proto.assay

    if track_proto.HasField('ontology_term'):
      track_metadata['ontology_curie'] = ontology.from_proto(
          track_proto.ontology_term
      ).ontology_curie

    if track_proto.HasField('biosample'):
      track_metadata['biosample_name'] = track_proto.biosample.name
      track_metadata['biosample_type'] = (
          dna_model_pb2.BiosampleType.Name(track_proto.biosample.type)
          .removeprefix('BIOSAMPLE_TYPE_')
          .lower()
      )
      if track_proto.biosample.HasField('stage'):
        track_metadata['biosample_life_stage'] = track_proto.biosample.stage

    if track_proto.HasField('transcription_factor_code'):
      track_metadata['transcription_factor'] = (
          track_proto.transcription_factor_code
      )

    if track_proto.HasField('histone_mark_code'):
      track_metadata['histone_mark'] = track_proto.histone_mark_code

    if track_proto.HasField('gtex_tissue'):
      track_metadata['gtex_tissue'] = track_proto.gtex_tissue

    if track_proto.HasField('data_source'):
      track_metadata['data_source'] = track_proto.data_source

    if track_proto.HasField('endedness'):
      track_metadata['endedness'] = (
          dna_model_pb2.Endedness.Name(track_proto.endedness)
          .removeprefix('ENDEDNESS_')
          .lower()
      )

    if track_proto.HasField('genetically_modified'):
      track_metadata['genetically_modified'] = track_proto.genetically_modified

    if track_proto.HasField('nonzero_mean'):
      track_metadata['nonzero_mean'] = track_proto.nonzero_mean

    metadata.append(track_metadata)
  if metadata:
    return pd.DataFrame(metadata)
  else:
    return pd.DataFrame(columns=['name', 'strand'])
